"""Test read arguments function for REANA templates."""


from unittest import TestCase

from benchtmpl.io.scanner import Scanner, ListReader
from benchtmpl.workflow.parameter.base import TemplateParameter
from benchtmpl.workflow.template.base import TemplateHandle

import benchtmpl.workflow.parameter.declaration as pd
import benchtmpl.workflow.template.util as tmpl


class TestReadTemplateArguments(TestCase):
    def test_read_with_record(self):
        """Read argument for a template that contains a parameter of data type
        DT_RECORD.
        """
        template = TemplateHandle(
            workflow_spec=dict(),
            parameters=[
                TemplateParameter(
                    pd.parameter_declaration(
                        identifier='codeFile',
                        data_type=pd.DT_FILE,
                        index=0,
                        default_value=None
                    )
                ),
                TemplateParameter(
                    pd.parameter_declaration(
                        identifier='sleeptime',
                        data_type=pd.DT_INTEGER,
                        index=1,
                        default_value=10
                    )
                ),
                TemplateParameter(
                    pd.parameter_declaration(
                        identifier='verbose',
                        data_type=pd.DT_BOOL,
                        index=2,
                        default_value=False
                    )
                ),
                TemplateParameter(
                    pd.parameter_declaration(
                        identifier='frac',
                        data_type=pd.DT_DECIMAL,
                        index=4
                    )
                ),
                TemplateParameter(
                    pd.parameter_declaration(
                        identifier='outputType',
                        index=3
                    )
                ),
            ]
        )
        sc = Scanner(reader=ListReader(['ABC.txt', 3, True, 'XYZ', 0.123]))
        arguments = tmpl.read(template.list_parameters(), scanner=sc)
        self.assertEqual(arguments['codeFile'].name, 'ABC.txt')
        self.assertEqual(arguments['sleeptime'], 3)
        self.assertEqual(arguments['verbose'], True)
        self.assertEqual(arguments['outputType'], 'XYZ')
        self.assertEqual(arguments['frac'], 0.123)


if __name__ == '__main__':
    import unittest
    unittest.main()



"""Test validation of template parameter declarations."""

from unittest import TestCase

from benchtmpl.workflow.parameter.base import TemplateParameter

import benchtmpl.error as err
import benchtmpl.workflow.parameter.declaration as pd


class TestParameterValidation(TestCase):
    def test_invalid_datatype(self):
        """Assert that an error is raised if a package declaration with an
        invalid data type is given.
        """
        # Ensure that it works with a valid data type
        p = pd.parameter_declaration(
            identifier='ABC',
            data_type=pd.DT_RECORD
        )
        with self.assertRaises(err.InvalidParameterError):
            p = pd.parameter_declaration(identifier='ABC', data_type='XYZ')
        # Specifying a non-string value should also raise an error
        with self.assertRaises(err.InvalidParameterError):
            pd.parameter_declaration(identifier='ABC', data_type=123)
        # Ensure that validation fails if data type is manipulated
        p = pd.parameter_declaration(
            identifier='ABC',
            data_type=pd.DT_RECORD
        )
        pd.validate_parameter(p)
        p[pd.LABEL_DATATYPE] = 'Something unknown'
        with self.assertRaises(err.InvalidParameterError):
            pd.validate_parameter(p)

    def test_invalid_identifier(self):
        """Assert that an error is raised if a package declaration with an
        invalid identifier is given.
        """
        # Ensure that it works with a valid identifier
        pd.parameter_declaration(identifier='ABC', data_type=pd.DT_BOOL)
        # Error is raised if identifier is None
        with self.assertRaises(err.InvalidParameterError):
            pd.parameter_declaration(identifier=None, data_type=pd.DT_BOOL)

    def test_maximal_declaration(self):
        """Test parameter declarations that provide values for all arguments.
        """
        # Set all parameter elements to values that are different from their
        # default value
        p = pd.parameter_declaration(
            identifier='ABC',
            name='XYZ',
            description='ABC to XYZ',
            data_type=pd.DT_INTEGER,
            index=10,
            required=False,
            values=[
                pd.enum_value(value=1),
                pd.enum_value(value=2, text='Two'),
                pd.enum_value(value=3, text='THREE', is_default=True)
            ],
            parent='DEF',
            default_value=5,
            as_const='data/names.txt'
        )
        self.assertTrue(isinstance(p, dict))
        self.assertEqual(p.get(pd.LABEL_ID), 'ABC')
        self.assertEqual(p.get(pd.LABEL_NAME), 'XYZ')
        self.assertEqual(p.get(pd.LABEL_DESCRIPTION), 'ABC to XYZ')
        self.assertEqual(p.get(pd.LABEL_DATATYPE), pd.DT_INTEGER)
        self.assertEqual(p.get(pd.LABEL_INDEX), 10)
        self.assertFalse(p.get(pd.LABEL_REQUIRED))
        self.assertEqual(p.get(pd.LABEL_PARENT), 'DEF')
        self.assertEqual(p.get(pd.LABEL_DEFAULT), 5)
        self.assertEqual(p.get(pd.LABEL_AS), 'data/names.txt')
        # Valudate value enumeration
        values = p.get(pd.LABEL_VALUES, [])
        self.assertEqual(len(values), 3)
        self.validate_value(values[0], 1, '1', False)
        self.validate_value(values[1], 2, 'Two', False)
        self.validate_value(values[2], 3, 'THREE', True)
        # Ensure that the returned dictionary is valid with respect to the
        # parameter schema declaration.
        pd.validate_parameter(p)

    def test_minimal_declaration(self):
        """Test parameter declarations that only provide the required arguments.
        """
        # Expect to get a dictionary that contains the identifier, name (both
        # equal to 'ABC'), a data type DT_STRING, an index of 0. The required
        #  flag is True.
        p = pd.parameter_declaration(identifier='ABC')
        self.assertTrue(isinstance(p, dict))
        self.assertEqual(p.get(pd.LABEL_ID), 'ABC')
        self.assertEqual(p.get(pd.LABEL_NAME), 'ABC')
        self.assertEqual(p.get(pd.LABEL_DESCRIPTION), 'ABC')
        self.assertEqual(p.get(pd.LABEL_DATATYPE), pd.DT_STRING)
        self.assertEqual(p.get(pd.LABEL_INDEX), 0)
        self.assertTrue(p.get(pd.LABEL_REQUIRED))
        # All other optional elements of the declaration are missing
        self.assertFalse(pd.LABEL_DEFAULT in p)
        self.assertFalse(pd.LABEL_PARENT in p)
        self.assertFalse(pd.LABEL_VALUES in p)
        # Ensure that the returned dictionary is valid with respect to the
        # parameter schema declaration.
        pd.validate_parameter(p)

    def test_parameter_declaration(self):
        """Test methods of the parameter declaration object."""
        obj = pd.set_defaults(pd.parameter_declaration(identifier='ABC'))
        obj[pd.LABEL_AS] = 'XYZ'
        p = TemplateParameter(obj)
        # Type flags
        self.assertFalse(p.is_bool())
        self.assertFalse(p.is_file())
        self.assertFalse(p.is_float())
        self.assertFalse(p.is_int())
        self.assertFalse(p.is_list())
        self.assertFalse(p.is_record())
        self.assertTrue(p.is_string())
        # Constant values
        self.assertTrue(p.has_constant())
        self.assertEqual(p.get_constant(), 'XYZ')
        # Error for invalid data type
        with self.assertRaises(err.InvalidParameterError):
            TemplateParameter({pd.LABEL_ID: 'A', pd.LABEL_DATATYPE: 'XYZ'})

    def test_set_defaults(self):
        """Test the set defaults method with minimal input."""
        p = pd.set_defaults({pd.LABEL_ID: 'My ID'})
        self.assertEqual(p.get(pd.LABEL_ID), 'My ID')
        self.assertEqual(p.get(pd.LABEL_NAME), 'My ID')
        self.assertEqual(p.get(pd.LABEL_DESCRIPTION), 'My ID')
        self.assertEqual(p.get(pd.LABEL_DATATYPE), pd.DT_STRING)
        self.assertEqual(p.get(pd.LABEL_INDEX), 0)
        self.assertTrue(p.get(pd.LABEL_REQUIRED))


    def test_validate_error(self):
        """Assert that errors are raised if an invalid parameter declaration is
        given to the validate_parameter function.
        """
        p = pd.parameter_declaration(identifier='ABC')
        # Ensure that creating a dictionary from a valid parameter declaration
        # is still valid
        pd.validate_parameter(dict(p))
        # Invalid data type for parameter identifier
        p_invalid = dict(p)
        p_invalid[pd.LABEL_ID] = 123
        with self.assertRaises(err.InvalidParameterError):
            pd.validate_parameter(p_invalid)
        # Invalid data type for parameter name
        p_invalid = dict(p)
        p_invalid[pd.LABEL_NAME] = 123
        with self.assertRaises(err.InvalidParameterError):
            pd.validate_parameter(p_invalid)
        # Invalid data type for parameter data type
        p_invalid = dict(p)
        p_invalid[pd.LABEL_DATATYPE] = 12.3
        with self.assertRaises(err.InvalidParameterError):
            pd.validate_parameter(p_invalid)
        # Invalid data type for parameter index
        p_invalid = dict(p)
        p_invalid[pd.LABEL_INDEX] = '12'
        with self.assertRaises(err.InvalidParameterError):
            pd.validate_parameter(p_invalid)
        # Invalid data type for parameter required
        p_invalid = dict(p)
        p_invalid[pd.LABEL_REQUIRED] = '12'
        with self.assertRaises(err.InvalidParameterError):
            pd.validate_parameter(p_invalid)

    def validate_value(self, obj, value, name, is_default):
        """Validate element in a parameter value enumeration."""
        self.assertEqual(obj[pd.LABEL_VALUE], value)
        self.assertEqual(obj[pd.LABEL_NAME], name)
        self.assertEqual(obj[pd.LABEL_IS_DEFAULT], is_default)


if __name__ == '__main__':
    import unittest
    unittest.main()



# -*- coding: utf-8 -*-

"""REANA-Demo-HelloWorld is a minimal demonstration of REANA system."""

from __future__ import absolute_import, print_function

import argparse
import errno
import os
import sys
import time


def hello(inputfile="inputs/names.txt", outputfile="outputs/greetings.txt", sleeptime=1.0):
    """Say 'Hello' to given name and store the greeting to a file.

    Writes the greeting character by character. An optional waiting period
    between writing of each character can be specified.

    :param inputfile: The relative path to the file containing the names
        of the persons to greet.
    :param outputfile: The relative path to the file where greeting
        should be stored. Creates the file if it does not exist.
    :param sleeptime: A waiting period (in seconds) between writing
        characters of greeting.

    """
    # detect names to greet:
    names = []
    with open(inputfile, 'r') as f:
        for line in f.readlines():
            names.append(line.strip())

    # ensure output directory exists:
    # influenced by http://stackoverflow.com/a/12517490
    if not os.path.exists(os.path.dirname(outputfile)):
        try:
            os.makedirs(os.path.dirname(outputfile))
        except OSError as exc:  # guard against race condition
            if exc.errno != errno.EEXIST:
                raise

    # write greetings:
    with open(outputfile, "at") as f:
        for name in names:
            message = "Hello " + name + "!\n"
            for char in message:
                f.write("{}".format(char))
            time.sleep(sleeptime)
    f.close()
    return


if __name__ == '__main__':
    args = sys.argv[1:]

    parser = argparse.ArgumentParser()

    parser.add_argument("-i", "--inputfile",
                        help="Relative path to the file containing \
                                  the names of the persons to greet. \n \
                                  [default=inputs/names.txt]",
                        default="inputs/names.txt",
                        required=False)

    parser.add_argument("-o", "--outputfile",
                        help="Relative path to the file where greeting \
                                  should be stored. \n \
                                  [default=outputs/greetings.txt]",
                        default="outputs/greetings.txt",
                        required=False)

    parser.add_argument("-s", "--sleeptime",
                        help="Waiting period (in seconds) between \
                                  writing characters of greeting. \n \
                                  [default=1]",
                        default=1.0,
                        type=float,
                        required=False)

    parsed_args = parser.parse_args(args)

    hello(parsed_args.inputfile, parsed_args.outputfile, parsed_args.sleeptime)



"""Test functionality of the template arguments modules."""

from unittest import TestCase


from benchtmpl.io.files.base import FileHandle
from benchtmpl.workflow.parameter.base import TemplateParameter
from benchtmpl.workflow.template.base import TemplateHandle
from benchtmpl.workflow.template.loader import DefaultTemplateLoader

import benchtmpl.workflow.parameter.declaration as pd
import benchtmpl.workflow.parameter.value as values
import benchtmpl.workflow.template.loader as tmpl


LOCAL_FILE = 'tests/files/schema.json'


class TestArgumentValues(TestCase):
    """Test parsing and validating argument values for parameterized workflow
    templates.
    """
    def test_flat_parse(self):
        """Test parsing arguments for a flat (un-nested) parameter declaration.
        """
        template = TemplateHandle(
            workflow_spec=dict(),
            parameters=[
                TemplateParameter(pd.parameter_declaration('A', data_type=pd.DT_INTEGER)),
                TemplateParameter(pd.parameter_declaration('B', data_type=pd.DT_BOOL)),
                TemplateParameter(pd.parameter_declaration('C', data_type=pd.DT_DECIMAL)),
                TemplateParameter(pd.parameter_declaration('D', data_type=pd.DT_FILE, required=False)),
                TemplateParameter(pd.parameter_declaration('E', data_type=pd.DT_STRING, required=False))
            ]
        )
        params = template.parameters
        fh = FileHandle(filepath=LOCAL_FILE)
        # Valid argument set
        args = values.parse_arguments(
            arguments={'A': 10, 'B': True, 'C': 12.5, 'D': fh, 'E': 'ABC'},
            parameters=params,
            validate=True
        )
        self.assertEqual(len(args), 5)
        for key in params.keys():
            self.assertTrue(key in args)
        values.parse_arguments(arguments=args, parameters=params, validate=False)
        # Error cases
        with self.assertRaises(ValueError):
            values.parse_arguments(arguments={'A': 10, 'Z': 0}, parameters=params)
        with self.assertRaises(ValueError):
            values.parse_arguments(arguments={'A': 10, 'B': True}, parameters=params)
        # Validate data type
        with self.assertRaises(ValueError):
            values.parse_arguments(
                arguments={'A': '10', 'B': True, 'C': 12.3, 'D': fh, 'E': 'ABC'},
                parameters=params,
                validate=True
            )
        with self.assertRaises(ValueError):
            values.parse_arguments(
                arguments={'A': 10, 'B': 23, 'C': 12.3, 'D': fh, 'E': 'ABC'},
                parameters=params,
                validate=True
            )
        with self.assertRaises(ValueError):
            values.parse_arguments(
                arguments={'A': 10, 'B': True, 'C': '12.3', 'D': fh, 'E': 'ABC'},
                parameters=params,
                validate=True
            )
        with self.assertRaises(ValueError):
            values.parse_arguments(
                arguments={'A': 10, 'B': True, 'C': 12.3, 'D': 'fh', 'E': 'ABC'},
                parameters=params,
                validate=True
            )
        with self.assertRaises(ValueError):
            values.parse_arguments(
                arguments={'A': 10, 'B': True, 'C': 12.3, 'D': fh, 'E': 12},
                parameters=params,
                validate=True
            )

    def test_nested_parse(self):
        """Test parsing arguments for a nested parameter declaration."""
        template = DefaultTemplateLoader().from_dict({
                tmpl.LABEL_WORKFLOW: dict(),
                tmpl.LABEL_PARAMETERS: [
                    pd.parameter_declaration('A', data_type=pd.DT_INTEGER),
                    pd.parameter_declaration('B', data_type=pd.DT_RECORD),
                    pd.parameter_declaration('C', data_type=pd.DT_DECIMAL, parent='B'),
                    pd.parameter_declaration('D', data_type=pd.DT_STRING, parent='B', required=False),
                    pd.parameter_declaration('E', data_type=pd.DT_LIST, required=False),
                    pd.parameter_declaration('F', data_type=pd.DT_INTEGER, parent='E'),
                    pd.parameter_declaration('G', data_type=pd.DT_DECIMAL,  parent='E', required=False)
                ]
            },
            validate=True
        )
        params = template.parameters
        # Without values for list parameters
        args = values.parse_arguments(
            arguments={'A': 10, 'B': {'C': 12.3}},
            parameters=params,
            validate=True
        )
        self.assertEqual(len(args), 2)
        self.assertTrue(args['B'].has('C'))
        self.assertFalse(args['B'].has('D'))
        self.assertEqual(args['B'].len(), 1)
        self.assertEqual(args['B'].get('C').value, 12.3)
        # With list arguments
        args = values.parse_arguments(
            arguments={
                'A': 10,
                'B': {'C': 12.3, 'D': 'ABC'},
                'E': [
                    {'F': 1},
                    {'F': 2},
                    {'F': 3, 'G': 0.9}
                ]
            },
            parameters=params,
            validate=True
        )
        self.assertEqual(len(args), 3)
        self.assertTrue(args['B'].has('C'))
        self.assertTrue(args['B'].has('D'))
        self.assertEqual(args['B'].len(), 2)
        self.assertEqual(args['E'].len(), 3)
        for arg in args['E'].value:
            if arg.get('F').value < 3:
                self.assertEqual(len(arg), 1)
            else:
                self.assertEqual(len(arg), 2)
        # Error cases
        with self.assertRaises(ValueError):
            values.parse_arguments(
                arguments={'A': 10, 'B': [{'C': 12.3}]},
                parameters=params,
                validate=True
            )

    def test_validate(self):
        """Test error cases for argument validation."""
        para_list = TemplateParameter(pd.parameter_declaration('E', data_type=pd.DT_LIST))
        values.TemplateArgument(parameter=para_list, value=1)
        with self.assertRaises(ValueError):
            values.TemplateArgument(parameter=para_list, value=1, validate=True)
        para_record = TemplateParameter(pd.parameter_declaration('E', data_type=pd.DT_RECORD))
        values.TemplateArgument(parameter=para_record, value=list())
        with self.assertRaises(ValueError):
            values.TemplateArgument(parameter=para_record, value=list(), validate=True)
        arg = values.TemplateArgument(
            parameter=TemplateParameter(pd.parameter_declaration('E', data_type=pd.DT_INTEGER)),
            value=1,
            validate=True
        )
        arg.data_type = 'unknown'
        with self.assertRaises(ValueError):
            arg.validate()


if __name__ == '__main__':
    import unittest
    unittest.main()



# This file is part of the Reproducible Open Benchmarks for Data Analysis
# Platform (ROB).
#
# Copyright (C) 2019 NYU.
#
# ROB is free software; you can redistribute it and/or modify it under the
# terms of the MIT License; see LICENSE file for more details.

"""Simple scanner classes to collect values (e.g., for template parameters)
from different sources (e.g. standard input). The Scanner class implements the
methods that are used to convert input values into the different data types
that are supported for parameter declarations.
"""

from abc import abstractmethod
from builtins import input


class Scanner(object):
    """Scanner that converts input tokens into the simple scalar data types
    boolean, float, file, integer and string that are supported in the template
    parameter declarations.
    """
    def __init__(self, reader=None):
        """Set reference to token reader. The token reader is for example
        helpful for unittest or if input is read from a file.

        By default the standard input reader is used.

        Parameters
        ----------
        reader: benchtmpl.io.scanner.TokenReader
            Reader for input tokens
        """
        self.reader = reader if not reader is None else InputReader()

    def next_bool(self, default_value=None):
        """Return next token as boolean. If the read token is an empty string
        the given default value is returned.

        Raises ValueError if the token value cannot be converted to boolean.

        Any of the following values will be recognized as valid boolean values.
        All values are case-insensitive:

        - True: [true, yes, y, t, 1]
        - False:[false, no, n, f, 0]

        Parameters
        ----------
        default_value: bool, optional
            Default value that is returned if the read token is an empty string

        Returns
        -------
        bool
        """
        val = self.reader.next_token()
        if val == '' and not default_value is None:
            return default_value
        if val.lower() in ['true', 'yes', 'y', 't', 1]:
            return True
        elif val.lower() in ['false', 'no', 'n', 'f', 0]:
            return False
        else:
            raise ValueError('not a boolean value \'' + str(val) + '\'')

    def next_file(self, default_value=None):
        """Return next token as string representing a file name. There are no
        tests performed to ensure whether the given value represents a valid
        path name or not (since the definition of a valid path name is very
        much dependent on the OS).

        Parameters
        ----------
        default_value: string, optional
            Default value that is returned if the read token is an empty string

        Returns
        -------
        string
        """
        val = self.reader.next_token().strip()
        if val == '' and not default_value is None:
            return default_value
        return val

    def next_float(self, default_value=None):
        """Return next token as float. Raises ValueError if the token value
        cannot be converted to float.

        Parameters
        ----------
        default_value: float, optional
            Default value that is returned if the read token is an empty string

        Returns
        -------
        float
        """
        val = self.reader.next_token()
        if val == '' and not default_value is None:
            return default_value
        return float(val)

    def next_int(self, default_value=None):
        """Return next token as integer. Raises ValueError if the token value
        cannot be converted to integer.

        Parameters
        ----------
        default_value: int, optional
            Default value that is returned if the read token is an empty string

        Returns
        -------
        int
        """
        val = self.reader.next_token()
        if val == '' and not default_value is None:
            return default_value
        return int(val)

    def next_string(self, default_value=None):
        """Return next token as string.

        Parameters
        ----------
        default_value: string, optional
            Default value that is returned if the read token is an empty string

        Returns
        -------
        string
        """
        val = self.reader.next_token()
        if val == '' and not default_value is None:
            return default_value
        return val


# ------------------------------------------------------------------------------
# Token Reader
# ------------------------------------------------------------------------------

class TokenReader(object):
    """Abstract token reader class that is used by the scanner to get the next
    input token.
    """
    @abstractmethod
    def next_token(self):
        """Read the next token.

        Returns
        -------
        string
        """
        raise NotImplementedError()


class InputReader(TokenReader):
    """Token reader that reads tokens from standard input."""
    def next_token(self):
        """Read token from standard input.

        Returns
        -------
        string
        """
        return input()


class ListReader(TokenReader):
    """Token reader that is initialized with a list of values. Returns tokens
    from the list until the end of the list is reached.
    """
    def __init__(self, tokens):
        """Initialize the list of tokens.

        Parameters
        ----------
        tokens: list(string)
            List of token values
        """
        self.tokens = tokens

    def next_token(self):
        """Return next token from the token list. If the end of the list has
        been reached None is returned.

        Returns
        -------
        string
        """
        if len(self.tokens) > 0:
            return str(self.tokens.pop(0))
        else:
            return None



# This file is part of the Reproducible Open Benchmarks for Data Analysis
# Platform (ROB).
#
# Copyright (C) 2019 NYU.
#
# ROB is free software; you can redistribute it and/or modify it under the
# terms of the MIT License; see LICENSE file for more details.

"""Definition of workflow states. The classes in this module represent the
different possible states of a workflow run. There are four different states:
(PENDING) the workflow run has been submitted and is waiting to start running,
(RUNNING) the workflow is actively executing at the moment,
(ERROR) workflow execution was interrupted by an error or canceled by the user,
(SUCCESS) the workflow run completed successfully.
"""

from datetime import datetime

import traceback


"""Definition of state type identifier."""
STATE_ERROR = 'ERROR'
STATE_PENDING = 'PENDING'
STATE_RUNNING = 'RUNNING'
STATE_SUCCESS = 'SUCCESS'


class WorkflowState(object):
    """The base class for workflow states contains the state type identifier
    that is used by the different state type methods. The state also maintains
    the timestamp of workflow run creation. Subclasses will add additional
    timestamps and properties.
    """
    def __init__(self, type_id, created_at=None):
        """Initialize the type identifier and the 'created at' timestamp.

        Parameters
        ----------
        type_id: string
            Type identifier
        """
        self.type_id = type_id
        self.created_at = created_at if not created_at is None else datetime.now()

    def is_active(self):
        """A workflow is in active state if it is either pending or running.

        Returns
        --------
        bool
        """
        return self.type_id in [STATE_PENDING, STATE_RUNNING]

    def is_error(self):
        """Returns True if the workflow state is of type ERROR.

        Returns
        -------
        bool
        """
        return self.type_id == STATE_ERROR

    def is_pending(self):
        """Returns True if the workflow state is of type PENDING.

        Returns
        -------
        bool
        """
        return self.type_id == STATE_PENDING

    def is_running(self):
        """Returns True if the workflow state is of type RUNNING.

        Returns
        -------
        bool
        """
        return self.type_id == STATE_RUNNING

    def is_success(self):
        """Returns True if the workflow state is of type SUCCESS.

        Returns
        -------
        bool
        """
        return self.type_id == STATE_SUCCESS


class StateError(WorkflowState):
    """Error state representation for a workflow run. The workflow has three
    timestamps: the workflow creation time, workflow run start time and the
    time at which the error occured (ot workflow was canceled). The state also
    maintains an optional list of error messages.
    """
    def __init__(self, created_at, started_at, stopped_at=None, messages=None):
        """Initialize the timestamps that are associated with the workflow
        state and the optional error messages.

        Parameters
        ----------
        created_at: datetime.datetime
            Timestamp of workflow creation
        started_at: datetime.datetime
            Timestamp when the workflow started running
        stopped_at: datetime.datetime, optional
            Timestamp when workflow error occurred or the when the workflow was
            canceled
        messages: list(string), optional
            Optional list of error messages
        """
        super(StateError, self).__init__(
            type_id=STATE_ERROR,
            created_at=created_at
        )
        self.started_at = started_at
        self.stopped_at = stopped_at if not stopped_at is None else datetime.now()
        self.messages = messages if not messages is None else list()


class StatePending(WorkflowState):
    """State representation for a pending workflow that is waiting to start
    running. The workflow has only one timestamp representing the workflow
    creation time.
    """
    def __init__(self, created_at=None):
        """Initialize the timestamp that is associated with the workflow state.

        Parameters
        ----------
        created_at: datetime.datetime, optional
            Timestamp of workflow creation
        """
        super(StatePending, self).__init__(
            type_id=STATE_PENDING,
            created_at=created_at
        )

    def start(self):
        """Get instance of running state with the same create at timestamp as
        this state and the started at with the current timestamp.

        Returns
        -------
        benchtmpl.workflow.state.StateRunning
        """
        return StateRunning(created_at=self.created_at)


class StateRunning(WorkflowState):
    """State representation for a active workflow run. The workflow has two
    timestamps: the workflow creation time and the workflow run start time.
    """
    def __init__(self, created_at, started_at=None):
        """Initialize the timestamps that are associated with the workflow
        state.

        Parameters
        ----------
        created_at: datetime.datetime
            Timestamp of workflow creation
        started_at: datetime.datetime
            Timestamp when the workflow started running
        """
        super(StateRunning, self).__init__(
            type_id=STATE_RUNNING,
            created_at=created_at
        )
        self.started_at = started_at if not started_at is None else datetime.now()

    def error(self, messages=None):
        """Get instance of error state for a running wokflow. If the exception
        that caused the workflow execution to terminate is given it will be used
        to create the list of error messages.

        Parameters
        ----------
        messages: list(string), optional
            Optional list of error messages

        Returns
        -------
        benchtmpl.workflow.state.StateError
        """
        return StateError(
            created_at=self.created_at,
            started_at=self.started_at,
            messages=messages
        )

    def success(self, resources=None):
        """Get instance of success state for a competed wokflow.

        Parameters
        ----------
        resources: dict(benchtmpl.workflow.resource.ResourceHandle), optional
            Optional dictionary of created resources

        Returns
        -------
        benchtmpl.workflow.state.StateSuccess
        """
        return StateSuccess(
            created_at=self.created_at,
            started_at=self.started_at,
            resources=resources
        )


class StateSuccess(WorkflowState):
    """Success state representation for a workflow run. The workflow has three
    timestamps: the workflow creation time, workflow run start time and the
    time when the workflow execution finished. The state also maintains handles
    to any resources that were created by the workflow run.
    """
    def __init__(self, created_at, started_at, finished_at=None, resources=None):
        """Initialize the timestamps that are associated with the workflow
        state and the set of created resources.

        Parameters
        ----------
        created_at: datetime.datetime
            Timestamp of workflow creation
        started_at: datetime.datetime
            Timestamp when the workflow started running
        finished_at: datetime.datetime, optional
            Timestamp when workflow execution completed
        resources: dict(benchtmpl.workflow.resource.ResourceHandle), optional
            Optional dictionary of created resources
        """
        super(StateSuccess, self).__init__(
            type_id=STATE_SUCCESS,
            created_at=created_at
        )
        self.started_at = started_at
        self.finished_at = finished_at if not finished_at is None else datetime.now()
        self.resources = resources if not resources is None else dict()



# This file is part of the Reproducible Open Benchmarks for Data Analysis
# Platform (ROB).
#
# Copyright (C) 2019 NYU.
#
# ROB is free software; you can redistribute it and/or modify it under the
# terms of the MIT License; see LICENSE file for more details.

"""Exceptions that are raised by the various components of the benchmark
templates.
"""


class TemplateError(Exception):
    """Base exception indicating that a component of benchmark templates
    encountered an error situation.
    """
    def __init__(self, message):
        """Initialize error message.

        Parameters
        ----------
        message : string
            Error message
        """
        Exception.__init__(self)
        self.message = message

    def __str__(self):
        """Get printable representation of the exception.

        Returns
        -------
        string
        """
        return self.message

# ------------------------------------------------------------------------------
# Backend
# ------------------------------------------------------------------------------

class MissingArgumentError(TemplateError):
    """Exception indicating that a required parameter in a workflow template
    has no argument given for a workflow run.
    """
    def __init__(self, identifier):
        """Initialize missing argument error message for parameter identifier.

        Parameters
        ----------
        identifier: string
            Unique parameter identifier
        """
        super(MissingArgumentError, self).__init__(
            message='missing argument for \'{}\''.format(identifier)
        )


class UnknownRunError(TemplateError):
    """Exception indicating that a given run identifier does not reference a
    known workflow run.
    """
    def __init__(self, identifier):
        """Initialize error message for unknown run identifier.

        Parameters
        ----------
        identifier: string
            Unique run identifier
        """
        super(UnknownRunError, self).__init__(
            message='unknown run \'{}\''.format(identifier)
        )


# ------------------------------------------------------------------------------
# Workflow Templates
# ------------------------------------------------------------------------------

class InvalidParameterError(TemplateError):
    """Exception indicating that a given template parameter is invalid.
    """
    def __init__(self, message):
        """Initialize error message. The message for invalid parameter errors
        is depending on the context

        Parameters
        ----------
        message : string
            Error message
        """
        super(InvalidParameterError, self).__init__(message=message)


class InvalidTemplateError(TemplateError):
    """Exception indicating that a given workflow template is invalid or has
    missing elements.
    """
    def __init__(self, message):
        """Initialize error message. The message for invalid template errors
        is depending on the context

        Parameters
        ----------
        message : string
            Error message
        """
        super(InvalidTemplateError, self).__init__(message=message)

class UnknownParameterError(TemplateError):
    """Exception indicating that a workflow specification references a parameter
    that is not defined for a given template.
    """
    def __init__(self, identifier):
        """Initialize error message for given parameter identifier.

        Parameters
        ----------
        identifier: string
            Unique template parameter identifier
        """
        super(UnknownParameterError, self).__init__(
            message='reference to undefined parameter \'{}\''.format(identifier)
        )


class UnknownTemplateError(TemplateError):
    """Exception indicating that a given template identifier does not reference
    an existing template.
    """
    def __init__(self, identifier):
        """Initialize error message for unknown template identifier.

        Parameters
        ----------
        identifier: string
            Unique template identifier
        """
        super(UnknownTemplateError, self).__init__(
            message='unknown template \'{}\''.format(identifier)
        )



# This file is part of the Reproducible Open Benchmarks for Data Analysis
# Platform (ROB).
#
# Copyright (C) 2019 NYU.
#
# ROB is free software; you can redistribute it and/or modify it under the
# terms of the MIT License; see LICENSE file for more details.

"""The template loader reads and writes template handles from and to files
on disk. This class implements a factory pattern for different variations of
workflow templates. Each variation of the template handle should provide an
implementation of the template loader. Thie loader is used by the template
repository to store and retrieve template specifications.

An implementation for the default base template handle is provided in this
module.
"""

from abc import abstractmethod

from benchtmpl.workflow.parameter.base import TemplateParameter

import benchtmpl.error as err
import benchtmpl.util.core as util
import benchtmpl.workflow.parameter.declaration as pd
import benchtmpl.workflow.parameter.util as putil
import benchtmpl.workflow.template.base as tmpl


"""Top-level elements of dictionary serialization for template handles."""
LABEL_ID = 'id'
LABEL_PARAMETERS = 'parameters'
LABEL_WORKFLOW = 'workflow'


class TemplateLoader(object):
    """The template loader is used by the template repository to store and
    retrieve workflow template specifications on disk.
    """
    @abstractmethod
    def from_dict(self, doc, identifier=None, base_dir=None, validate=True):
        """Create an instance of the template handle for a dictionary
        serialization. The elements in the dictionary and their structure is
        expected to the elements and structure that is produced by the to_dict()
        method.

        An InvalidTemplateError is raised if the dictionary does not represent
        a valid serialization of a workflow handle.

        If the valid flag is True all given template parameter declarations are
        validated against the parameter schema. An InvalidTemplateError is
        raised if (i) any of the given parameter declarations fails the
        validation, or (ii) the workflow specifiction references parameters for
        which no declaration is given.

        Parameters
        ----------
        dict: dict
            Dictionary serialization of a workflow template
        identifier: string, optional
            Unique template identifier. This value will override the value in
            the document.
        base_dir: string, optional
            Optional path to directory on disk that contains static files that
            are required to run the represented workflow. This value will
            override the value in the document.
        validate: bool, optional
            Flag indicating if given template parameter declarations are to be
            validated against the parameter schema or not.

        Returns
        -------
        benchtmpl.workflow.template.base.TemplateHandle

        Raises
        ------
        benchtmpl.error.InvalidTemplateError
        benchtmpl.error.UnknownParameterError
        """
        raise NotImplementedError()

    def load(self, filename, identifier=None, base_dir=None, format=None, validate=True):
        """Create an instance of the template handle for a serialization if the
        template handle that is stored in a file.

        Parameters
        ----------
        filename: string
            Path to a file on disk
        identifier: string, optional
            Unique template identifier. This value will override the value in
            the document.
        base_dir: string, optional
            Optional path to directory on disk that contains static files that
            are required to run the represented workflow. This value will
            override the value in the document.
        format: string, optional
            Optional file format identifier. The default is YAML
        validate: bool, optional
            Flag indicating if given template parameter declarations are to be
            validated against the parameter schema or not.

        Returns
        -------
        benchtmpl.workflow.template.base.TemplateHandle

        Raises
        ------
        benchtmpl.error.InvalidTemplateError
        """
        doc = util.read_object(filename, format=format)
        return self.from_dict(
            doc,
            identifier=identifier,
            base_dir=base_dir,
            validate=validate
        )

    @abstractmethod
    def to_dict(self, template):
        """Get a dictionary serialization for the workflow template.

        Parameters
        ----------
        template: benchtmpl.workflow.template.base.TemplateHandle
            Instance of a template handle class that is supported by the
            template loader

        Returns
        -------
        dict
        """
        raise NotImplementedError()

    def write(self, template, filename, format=None):
        """Write the given template handle to disk.

        Parameters
        ----------
        template: benchtmpl.workflow.template.base.TemplateHandle
            Instance of a template handle class that is supported by the
            template loader
        filename: string
            Path to a file to which the serialized handle is written
        format: string, optional
            Optional file format identifier. The default is YAML

        """
        util.write_object(
            obj=self.to_dict(template),
            filename=filename
        )


class DefaultTemplateLoader(TemplateLoader):
    """Default implementation of the loader class for the basic template handle.
    This is the default loader that is used by the template repository.
    """
    def from_dict(self, doc, identifier=None, base_dir=None, validate=True):
        """Create an instance of the template handle for a dictionary
        serialization. The expected three top-level elements of the dictionary
        are 'workflow', 'id', and 'parameters'. The last two elements are
        optional.

        Parameters
        ----------
        dict: dict
            Dictionary serialization of a workflow template
        identifier: string, optional
            Unique template identifier. This value will override the value in
            the document.
        base_dir: string, optional
            Optional path to directory on disk that contains static files that
            are required to run the represented workflow. This value will
            override the value in the document.
        validate: bool, optional
            Flag indicating if given template parameter declarations are to be
            validated against the parameter schema or not.

        Returns
        -------
        benchtmpl.workflow.template.base.TemplateHandle

        Raises
        ------
        benchtmpl.error.InvalidTemplateError
        benchtmpl.error.UnknownParameterError
        """
        # Ensure that the mandatory elements are present
        if not LABEL_WORKFLOW in doc:
            raise err.InvalidTemplateError('missing element \'{}\''.format(LABEL_WORKFLOW))
        workflow_spec = doc[LABEL_WORKFLOW]
        # Add given parameter declarations to the parameter list. Ensure that
        # all default values are set
        parameters = dict()
        if LABEL_PARAMETERS in doc:
            parameters = putil.create_parameter_index(
                doc[LABEL_PARAMETERS],
                validate=validate
            )
        # Ensure that the workflow specification does not reference undefined
        # parameters if validate flag is True.
        if validate:
            for key in tmpl.get_parameter_references(workflow_spec):
                if not key in parameters:
                    raise err.UnknownParameterError(key)
        # Get identifier if present in document
        if LABEL_ID in doc:
            identifier = doc[LABEL_ID]
        return tmpl.TemplateHandle(
            identifier=identifier,
            base_dir=base_dir,
            workflow_spec=workflow_spec,
            parameters=list(parameters.values())
        )

    def to_dict(self, template):
        """Get dictionary serializationfor the wrokflow template.

        Parameters
        ----------
        template: benchtmpl.workflow.template.base.TemplateHandle
            Instance of a template handle class that is supported by the
            template loader

        Returns
        -------
        dict
        """
        return {
            LABEL_ID: template.identifier,
            LABEL_WORKFLOW: template.workflow_spec,
            LABEL_PARAMETERS: [p.to_dict() for p in template.parameters.values()]
        }



# This file is part of the Reproducible Open Benchmarks for Data Analysis
# Platform (ROB).
#
# Copyright (C) 2019 NYU.
#
# ROB is free software; you can redistribute it and/or modify it under the
# terms of the MIT License; see LICENSE file for more details.

"""The template argument class is used to represent user-provided values for
workflow template parameters. The class is a simple wrapper that combines
the value and the meta-data in the parameter declaration.
"""

from benchtmpl.io.files.base import FileHandle
from benchtmpl.workflow.parameter.base import ParameterBase

import benchtmpl.workflow.parameter.declaration as pd


class TemplateArgument(ParameterBase):
    """Template arguments capture user-provided values for workflow template
    parameters that are used to instantiate and execute a parameterized workflow
    specification. The argument class captures the actual value and provides
    access to the parameter meta-data.
    """
    def __init__(self, parameter, value, validate=False):
        """Initialize the parameter value and meta-data. The type of the value
        argument depends on the data type of the parameter. If the parameter is
        of type record the value is expected to be a dictionary of arguments.
        If the parameter is of type list the value is expected to be a list of
        lists.

        Parameters
        ----------
        parameter: benchtmpl.workflow.parameter.base.TemplateParameter
            Parameter declaration
        value: list or dict or scalar or benchtmpl.io.files.base.FileHandle
            Parameter value. the type depends on the parameter data type.
        validate: bool, optional
            Validate the argument value against the parameter declaration if
            set to True

        Raises
        ------
        ValueError
        """
        super(TemplateArgument, self).__init__(
            identifier=parameter.identifier,
            data_type=parameter.data_type
        )
        self.parameter = parameter
        self.value = value
        # Validate the argument value if the validate flag is set to True
        if validate:
            self.validate()

    def get(self, key):
        """Shortcut to get element for parameters of type record or list.

        Parameters
        ----------
        key: string or int
            Depending on the parameter type the key is either a list index or
            a dictionary key

        Returns
        -------
        benchtmpl.workflow.parameter.value.TemplateArgument
        """
        return self.value[key]

    def has(self, key):
        """Test if the argument value for a record parameter contains a value
        for the given record element.

        Parameters
        ----------
        key: string
            Unique identifier for child element in a parameter record

        Returns
        -------
        bool
        """
        return key in self.value

    def len(self):
        """Shortcut to get length of list if parameter is of type list.

        Returns
        -------
        int
        """
        return len(self.value)

    def validate(self):
        """Validate the argument value against the parameter declaration. This
        method does not return any value but it raises a ValueError if the
        argument value is not valid with respect to the parameter declaration.

        Raises
        ------
        ValueError
        """
        if self.is_bool():
            if not isinstance(self.value, bool):
                raise ValueError('expected bool for \'{}\''.format(self.identifier))
        elif self.is_float():
            if not isinstance(self.value, float):
                raise ValueError('expected float for \'{}\''.format(self.identifier))
        elif self.is_int():
            if not isinstance(self.value, int):
                raise ValueError('expected int for \'{}\''.format(self.identifier))
        elif self.is_string():
            if not isinstance(self.value, str):
                raise ValueError('expected string for \'{}\''.format(self.identifier))
        elif self.is_file():
            # Expects a file handle
            if not isinstance(self.value, FileHandle):
                raise ValueError('expected file handle for \'{}\''.format(self.identifier))
        elif self.is_list():
            if not isinstance(self.value, list):
                raise ValueError('expected list for \'{}\''.format(self.identifier))
            for record in self.value:
                for arg in record.values():
                    arg.validate()
        elif self.is_record():
            if not isinstance(self.value, dict):
                raise ValueError('expected dictionary for \'{}\''.format(self.identifier))
            for arg in self.value.values():
                arg.validate()
        else:
            raise ValueError('unknown data type \'{}\''.format(self.data_type))


# ------------------------------------------------------------------------------
# Helper Methods
# ------------------------------------------------------------------------------

def mandatory_arguments(parameters, parent=None):
    """Get a list of parameter names that are mandatory. The optional parent
    parameter allows to request mandatory parameter for nested components.

    Parameters
    ----------
    parameters: dict(benchtmpl.workflow.parameter.base.TemplateParameter)
        Dictionary of parameter declarations
    parent: benchtmpl.workflow.parameter.base.TemplateParameter
        Parent paremeter declaration for nested structures

    Returns
    -------
    list(string)
    """
    result = list()
    for para in parameters.values():
        if not para.is_required:
            continue
        if para.parent == parent:
            result.append(para.identifier)
    return result


def parse_arguments(arguments, parameters, validate=False, parent=None):
    """Convert a dictionary of argument identifier and argument value pairs into
    a dictionary of template argument instances.

    Parameters
    ----------
    arguments: dict()
        Key, value pairs of argument identifier and argument value
    parameters: dict(benchtmpl.workflow.parameter.base.TemplateParameter)
        Dictionary of parameter declarations
    validate: bool, optional
        Validate argument value agains parameter declaration if True
    parent: benchtmpl.workflow.parameter.base.TemplateParameter
        Parent paremeter declaration for nested structures

    Returns
    -------
    dict(benchtmpl.workflow.parameter.value.TemplateArgument)

    Raises
    ------
    ValueError
    """
    result = dict()
    for arg_id, arg_value in arguments.items():
        # Get the parameter declaration. Raise error if the parameter id is
        # unknown.
        if not arg_id in parameters:
            raise ValueError('unknown argument \'{}\''.format(arg_id))
        para = parameters[arg_id]
        if isinstance(arg_value, list) and para.is_list():
            # Expects a list of records
            value = list()
            for rec in arg_value:
                value.append(
                    parse_arguments(
                        arguments=rec,
                        parameters=parameters,
                        validate=validate,
                        parent=para
                    )
                )
        elif isinstance(arg_value, dict) and para.is_record():
            # Expects a dictionary of argument values (i.e., a record)
            value = parse_arguments(
                arguments=arg_value,
                parameters=parameters,
                validate=validate,
                parent=para
            )
        elif not (para.is_list() or para.is_record()):
            value = arg_value
        else:
            raise ValueError('invalid value for \'{}\''.format(arg_id))
        result[arg_id] = TemplateArgument(
            parameter=para,
            value=value,
            validate=validate
        )
    # Ensure that all mandatory arguments are given
    for key in mandatory_arguments(parameters, parent=parent):
        if not key in result:
            raise ValueError('missing value for \'{}\''.format(key))
    return result



"""This is an adopted version of the REANA-Demo-HelloWorld demo. The main
intend is for testing of the simple multi process backend.
"""

from __future__ import absolute_import, print_function

import argparse
import errno
import os
import sys
import time


def hello(inputfile, outputfile, greeting='Hello', sleeptime=0.0):
    """Write greeting for every name in a given input file to the output file.
    The optional waiting period delays the output between each input name.

    """
    # detect names to greet:
    names = []
    with open(inputfile, 'r') as f:
        for line in f.readlines():
            names.append(line.strip())

    # ensure output directory exists:
    # influenced by http://stackoverflow.com/a/12517490
    if not os.path.exists(os.path.dirname(outputfile)):
        try:
            os.makedirs(os.path.dirname(outputfile))
        except OSError as exc:  # guard against race condition
            if exc.errno != errno.EEXIST:
                raise

    # write greetings:
    with open(outputfile, "at") as f:
        for name in names:
            f.write(greeting + " " + name + "!\n")
            f.flush()
            time.sleep(sleeptime)


if __name__ == '__main__':
    args = sys.argv[1:]

    parser = argparse.ArgumentParser()
    parser.add_argument("-i", "--inputfile", required=True)
    parser.add_argument("-o", "--outputfile", required=True)
    parser.add_argument("-g", "--greeting", default='Hello', required=False)
    parser.add_argument("-s", "--sleeptime", default=1.0, type=float, required=False)

    parsed_args = parser.parse_args(args)

    hello(
        inputfile=parsed_args.inputfile,
        outputfile=parsed_args.outputfile,
        greeting=parsed_args.greeting,
        sleeptime=parsed_args.sleeptime
    )



"""Test basic token scanner functionality to collect template parameter values
from standard input.
"""

from unittest import TestCase

from benchtmpl.io.scanner import Scanner, ListReader
from benchtmpl.workflow.parameter.base import TemplateParameter

import benchtmpl.workflow.parameter.declaration as pd


class TestScanner(TestCase):
    def test_propmpt(self):
        """Test generated prompts when reading parameter values from standard
        input.
        """
        # BOOL
        p = TemplateParameter(pd.set_defaults(pd.parameter_declaration(
            identifier='ABC',
            data_type=pd.DT_BOOL
        ))).prompt()
        self.assertEqual(p, 'ABC (bool): ')
        # FILE
        p = TemplateParameter(pd.set_defaults(pd.parameter_declaration(
            identifier='ABC',
            data_type=pd.DT_FILE
        ))).prompt()
        self.assertEqual(p, 'ABC (file): ')
        # FLOAT
        p = TemplateParameter(pd.set_defaults(pd.parameter_declaration(
            identifier='ABC',
            data_type=pd.DT_DECIMAL
        ))).prompt()
        self.assertEqual(p, 'ABC (decimal): ')
        # INTEGER
        p = TemplateParameter(pd.set_defaults(pd.parameter_declaration(
            identifier='ABC',
            data_type=pd.DT_INTEGER
        ))).prompt()
        self.assertEqual(p, 'ABC (integer): ')
        # STRING
        p = TemplateParameter(pd.set_defaults(pd.parameter_declaration(
            identifier='ABC',
            data_type=pd.DT_STRING
        ))).prompt()
        self.assertEqual(p, 'ABC (string): ')
        # Default values in prompts
        p = TemplateParameter(pd.set_defaults(pd.parameter_declaration(
            identifier='ABC',
            data_type=pd.DT_INTEGER,
            default_value=100
        ))).prompt()
        self.assertEqual(p, 'ABC (integer) [default 100]: ')
        p = TemplateParameter(pd.set_defaults(pd.parameter_declaration(
            identifier='ABC',
            data_type=pd.DT_STRING,
            default_value=100
        ))).prompt()
        self.assertEqual(p, 'ABC (string) [default \'100\']: ')

    def test_scan_default_values(self):
        """Test return of default values when reading empty input."""
        sc = Scanner(reader=ListReader(5 * ['']))
        self.assertEqual(sc.next_int(default_value=11), 11)
        self.assertEqual(sc.next_float(default_value=1.23), 1.23)
        self.assertFalse(sc.next_bool(default_value=False))
        self.assertEqual(sc.next_file(default_value='file.txt'), 'file.txt')
        self.assertEqual(sc.next_string(default_value='Default text'), 'Default text')

    def test_scan_scalar_values(self):
        """Test parsing of scalar tokens."""
        sc = Scanner(reader=ListReader(['3', '34.56', 'FALSE', 'y', 'data/names.txt', 'Some text']))
        self.assertEqual(sc.next_int(), 3)
        self.assertEqual(sc.next_float(), 34.56)
        self.assertFalse(sc.next_bool())
        self.assertTrue(sc.next_bool())
        self.assertEqual(sc.next_file(), 'data/names.txt')
        self.assertEqual(sc.next_string(), 'Some text')
        # Value errors when parsing invalid tokens
        sc = Scanner(reader=ListReader(['3', 'FALSE', 'data/names.txt']))
        with self.assertRaises(ValueError):
            sc.next_bool()
        with self.assertRaises(ValueError):
            sc.next_int()
        with self.assertRaises(ValueError):
            sc.next_float()


if __name__ == '__main__':
    import unittest
    unittest.main()



# This file is part of the Reproducible Open Benchmarks for Data Analysis
# Platform (ROB).
#
# Copyright (C) 2019 NYU.
#
# ROB is free software; you can redistribute it and/or modify it under the
# terms of the MIT License; see LICENSE file for more details.

"""Helper methods for workflow template parameters."""


from benchtmpl.workflow.parameter.base import TemplateParameter

import benchtmpl.error as err
import benchtmpl.workflow.parameter.declaration as pd


def create_parameter_index(parameters, validate=True):
    """Create instances of template parameters from a list of dictionaries
    containing parameter declarations. The result is a dictionary containing the
    top-level parameters, indexed by their unique identifier.

    Parameters
    ----------
    parameters: list(dict)
        List of dictionaries containing template parameter declarations
    validate: bool, optional
        Flag indicating if given template parameter declarations are to be
        validated against the parameter schema or not.

    Returns
    -------
    dict(benchtmpl.workflow.parameter.base.TemplateParameter)

    Raises
    ------
    benchtmpl.error.InvalidTemplateError
    benchtmpl.error.UnknownParameterError
    """
    result = dict()
    for para in parameters:
        # Validate the template parameters if the validate flag is True
        if validate:
            pd.validate_parameter(para)
        # Create a TemplateParameter instance for the parameter. Keep
        # track of children for parameter that are of type DT_LIST or
        # DT_RECORD. Children are added after all parameters have been
        # instantiated.
        p_id = para[pd.LABEL_ID]
        # Ensure that the identifier of all parameters are unique
        if p_id in result:
            raise err.InvalidTemplateError('parameter \'{}\' not unique'.format(p_id))
        c = None
        if para[pd.LABEL_DATATYPE] in [pd.DT_LIST, pd.DT_RECORD]:
            c = list()
        tp = TemplateParameter(pd.set_defaults(para), children=c)
        result[p_id] = tp
    # Add parameter templates to the list of children for their
    # respective parent (if given). We currently only support one level
    # of nesting.
    for para in parameters:
        if pd.LABEL_PARENT in para:
            p_id = para[pd.LABEL_ID]
            parent = para[pd.LABEL_PARENT]
            if not parent is None:
                result[parent].add_child(result[p_id])
    return result


def sort_parameters(parameters):
    """Sort a given list of parameter declaration by the parameter index that is
    part of the parameter declaration. Parameters with the same index value are
    ordered by increasing value of their name.

    Parameters
    ----------
    parameters: list(benchtmpl.workflow.parameter.base.TemplateParameter)
        List of termplate parameters

    Returns
    -------
    list(benchtmpl.workflow.parameter.base.TemplateParameter)
    """
    return sorted(parameters, key=lambda p: (p.index, p.identifier))



"""Test functionality of the template for benchmark specifications and the
template loader.
"""

from unittest import TestCase

from benchtmpl.workflow.benchmark.loader import BenchmarkTemplateLoader

import benchtmpl.error as err


TEMPLATE_FILE_1 = './tests/files/benchmark/template_1.yaml'
TEMPLATE_FILE_ERR_1 = './tests/files/benchmark/template_2.yaml'
TEMPLATE_FILE_ERR_2 = './tests/files/benchmark/template_3.yaml'
TEMPLATE_FILE_ERR_3 = './tests/files/template/template.yaml'


class TestBenchmarkLoader(TestCase):
    """Test benchmark template serialization and the template loader."""
    def test_load_from_file(self):
        """Test loading benchmark templates from a valid and invalid template
        files.
        """
        loader = BenchmarkTemplateLoader()
        template = loader.load(TEMPLATE_FILE_1)
        self.assertEqual(len(template.parameters), 3)
        for key in ['names', 'sleeptime', 'greeting']:
            self.assertTrue(key in template.parameters)
        self.assertTrue(len(template.schema.columns), 3)
        # Test error cases
        with self.assertRaises(err.InvalidTemplateError):
            loader.load(TEMPLATE_FILE_ERR_1)
        with self.assertRaises(err.InvalidTemplateError):
            loader.load(TEMPLATE_FILE_ERR_2)
        # A plain template file without schema information will also raise
        # an error when using the benchmark loader
        with self.assertRaises(err.InvalidTemplateError):
            loader.load(TEMPLATE_FILE_ERR_3)

    def test_template_serialization(self):
        """Test template serialization."""
        loader = BenchmarkTemplateLoader()
        template = loader.load(TEMPLATE_FILE_1)
        tmpl_ser = loader.from_dict(
            loader.to_dict(template),
            identifier=template.identifier,
            base_dir=template.base_dir
        )
        self.assertEqual(template.identifier, tmpl_ser.identifier)
        self.assertEqual(template.base_dir, tmpl_ser.base_dir)
        self.assertEqual(len(tmpl_ser.parameters), 3)
        for key in ['names', 'sleeptime', 'greeting']:
            self.assertTrue(key in tmpl_ser.parameters)
        self.assertTrue(len(tmpl_ser.schema.columns), 3)


if __name__ == '__main__':
    import unittest
    unittest.main()



"""Test benchmark result schema objects."""

from unittest import TestCase

import benchtmpl.workflow.benchmark.schema as schema

import benchtmpl.workflow.parameter.declaration as pd


class TestBenchmarkResultSchema(TestCase):
    def test_column_serialization(self):
        """Test serialization of column objects."""
        # Ensure the required default value is set properly
        col = schema.BenchmarkResultColumn(
            identifier='col_1',
            name='Column 1',
            data_type=pd.DT_INTEGER
        )
        self.assertEqual(col.identifier, 'col_1')
        self.assertEqual(col.name, 'Column 1')
        self.assertEqual(col.data_type, pd.DT_INTEGER)
        self.assertTrue(col.required)
        self.assertEqual(type(col.required), bool)
        self.assertFalse(col.is_default)
        self.assertTrue(col.is_desc())
        col = schema.BenchmarkResultColumn.from_dict(col.to_dict())
        self.assertEqual(col.identifier, 'col_1')
        self.assertEqual(col.name, 'Column 1')
        self.assertEqual(col.data_type, pd.DT_INTEGER)
        self.assertTrue(col.required)
        self.assertEqual(type(col.required), bool)
        self.assertFalse(col.is_default)
        self.assertTrue(col.is_desc())
        # Test serialization if required value is given
        col = schema.BenchmarkResultColumn(
            identifier='col_1',
            name='Column 1',
            data_type=pd.DT_INTEGER,
            required=False,
            is_default=True,
            sort_order='ASC'
        )
        self.assertEqual(col.identifier, 'col_1')
        self.assertEqual(col.name, 'Column 1')
        self.assertEqual(col.data_type, pd.DT_INTEGER)
        self.assertFalse(col.required)
        self.assertEqual(type(col.required), bool)
        self.assertTrue(col.is_default)
        self.assertFalse(col.is_desc())
        col = schema.BenchmarkResultColumn.from_dict(col.to_dict())
        self.assertEqual(col.identifier, 'col_1')
        self.assertEqual(col.name, 'Column 1')
        self.assertEqual(col.data_type, pd.DT_INTEGER)
        self.assertFalse(col.required)
        self.assertEqual(type(col.required), bool)
        self.assertTrue(col.is_default)
        self.assertFalse(col.is_desc())

    def test_schema_serialization(self):
        """Test creating schema objects from dictionaries and vice versa."""
        s = schema.BenchmarkResultSchema.from_dict({
            schema.LABEL_RESULT_FILE: 'results.json',
            schema.LABEL_SCHEMA: [
                schema.BenchmarkResultColumn(
                    identifier='col_1',
                    name='Column 1',
                    data_type=pd.DT_INTEGER
                ).to_dict(),
                schema.BenchmarkResultColumn(
                    identifier='col_2',
                    name='Column 2',
                    data_type=pd.DT_DECIMAL,
                    required=True
                ).to_dict(),
                schema.BenchmarkResultColumn(
                    identifier='col_3',
                    name='Column 3',
                    data_type=pd.DT_STRING,
                    required=False
                ).to_dict()
            ]
        })
        self.validate_schema(s)
        # Recreate the object from its serialization
        s = schema.BenchmarkResultSchema.from_dict(s.to_dict())
        self.validate_schema(s)
        # Error cases
        with self.assertRaises(ValueError):
            doc = s.to_dict()
            doc['unknown'] = 'A'
            schema.BenchmarkResultSchema.from_dict(doc)
        with self.assertRaises(ValueError):
            doc = s.to_dict()
            del doc[schema.LABEL_RESULT_FILE]
            schema.BenchmarkResultSchema.from_dict(doc)
        with self.assertRaises(ValueError):
            schema.BenchmarkResultColumn(
                identifier='col_1',
                name='Column 1',
                data_type=pd.DT_LIST
            )

    def validate_column(self, column, identifier, name, data_type, required):
        """Ensure that the given column matches the respective arguments."""
        self.assertEqual(column.identifier, identifier)
        self.assertEqual(column.name, name)
        self.assertEqual(column.data_type, data_type)
        self.assertEqual(column.required, required)

    def validate_schema(self, schema):
        """Validate that the given schema."""
        self.assertEqual(len(schema.columns), 3)
        self.validate_column(
            column=schema.columns[0],
            identifier='col_1',
            name='Column 1',
            data_type=pd.DT_INTEGER,
            required=True
        )
        self.validate_column(
            column=schema.columns[1],
            identifier='col_2',
            name='Column 2',
            data_type=pd.DT_DECIMAL,
            required=True
        )
        self.validate_column(
            column=schema.columns[2],
            identifier='col_3',
            name='Column 3',
            data_type=pd.DT_STRING,
            required=False
        )


if __name__ == '__main__':
    import unittest
    unittest.main()



# This file is part of the Reproducible Open Benchmarks for Data Analysis
# Platform (ROB).
#
# Copyright (C) 2019 NYU.
#
# ROB is free software; you can redistribute it and/or modify it under the
# terms of the MIT License; see LICENSE file for more details.

"""Helper methods to read argument values for template parameters from standard
input. These methods are primarily intended to be used by a command line version
of the reproducible benchmark engine.
"""

from __future__ import print_function

from benchtmpl.io.files.base import FileHandle
from benchtmpl.io.scanner import Scanner



def read(parameters, scanner=None):
    """Read values for each of the template parameters using a given input
    scanner. If no scanner is given values are read from standard input.

    Returns a dictionary of argument values that can be passed to the workflow
    execution engine to run a parameterized workflow.

    Parameters
    ----------
    parameters: list(benchtmpl.workflow.parameter.base.TemplateParameter)
        List of workflow template parameter declarations
    scanner: benchtmpl.io.scanner.Scanner
        Input scanner to read parameter values

    Returns
    -------
    dict
    """
    sc = scanner if not scanner is None else Scanner()
    arguments = dict()
    for para in parameters:
        # Skip nested parameter
        if not para.parent is None:
            continue
        if para.is_list():
            raise ValueError('lists are not supported yet')
        elif para.is_record():
            # A record can only appear once and all record children have
            # global unique identifier. Thus, we can add values for each
            # of the children directly to the arguments dictionary
            for child in para.children:
                val = read_parameter(child, sc, prompt_prefix='  ')
                if not val is None:
                    arguments[child.identifier] = val
        else:
            val = read_parameter(para, sc)
            if not val is None:
                arguments[para.identifier] = val
    return arguments



def read_parameter(para, scanner, prompt_prefix=''):
    """Read value for a given template parameter declaration. Prompts the
    user to enter a value for the given parameter and returns the converted
    value that was entered by the user.

    Parameters
    ----------
    para: benchtmpl.workflow.parameter.TemplateParameter
        Workflow template parameter declaration

    Returns
    -------
    bool or float or int or string or list
    """
    done = False
    while not done:
        done = True
        print(prompt_prefix + para.prompt(), end='')
        try:
            if para.is_bool():
                return scanner.next_bool(default_value=para.default_value)
            elif para.is_file():
                filename = scanner.next_file(default_value=para.default_value)
                return FileHandle(filepath=filename)
            elif para.is_float():
                return scanner.next_float(default_value=para.default_value)
            elif para.is_int():
                return scanner.next_int(default_value=para.default_value)
            else:
                return scanner.next_string(default_value=para.default_value)
        except ValueError as ex:
            print(ex)
            done = False



"""Simple tests to ensure correctness of the workflow state classes."""

import datetime as dt

from unittest import TestCase

from benchtmpl.workflow.state import StateError, StatePending, StateRunning, StateSuccess
from benchtmpl.workflow.resource.base import FileResource


LOCAL_FILE = 'tests/files/schema.json'


class TestWorkflowStates(TestCase):
    """Test instantiating the different workflow state classes."""
    def test_error_state(self):
        """Test creating instances of the error state class."""
        created_at = dt.datetime.now()
        started_at = created_at + dt.timedelta(seconds=10)
        stopped_at = started_at + dt.timedelta(seconds=10)
        state = StateError(
            created_at=created_at,
            started_at=started_at,
            stopped_at=stopped_at
        )
        self.assertTrue(state.is_error())
        self.assertFalse(state.is_pending())
        self.assertFalse(state.is_running())
        self.assertFalse(state.is_success())
        self.assertFalse(state.is_active())
        self.assertEqual(state.created_at, created_at)
        self.assertEqual(state.started_at, started_at)
        self.assertEqual(state.stopped_at, stopped_at)
        self.assertEqual(len(state.messages), 0)
        state = StateError(
            created_at=created_at,
            started_at=started_at,
            stopped_at=stopped_at,
            messages=['A', 'B', 'C']
        )
        self.assertEqual(state.created_at, created_at)
        self.assertEqual(state.started_at, started_at)
        self.assertEqual(state.stopped_at, stopped_at)
        self.assertEqual(len(state.messages), 3)

    def test_pending_state(self):
        """Test creating instances of the pending state class."""
        created_at = dt.datetime.now()
        state = StatePending(created_at)
        self.assertTrue(state.is_pending())
        self.assertTrue(state.is_active())
        self.assertFalse(state.is_error())
        self.assertFalse(state.is_running())
        self.assertFalse(state.is_success())
        self.assertEqual(state.created_at, created_at)
        running = state.start()
        self.assertEqual(state.created_at, running.created_at)
        self.assertIsNotNone(running.started_at)

    def test_running_state(self):
        """Test creating instances of the running state class."""
        created_at = dt.datetime.now()
        started_at = created_at + dt.timedelta(seconds=10)
        state = StateRunning(created_at=created_at, started_at=started_at)
        self.assertTrue(state.is_active())
        self.assertTrue(state.is_running())
        self.assertFalse(state.is_pending())
        self.assertFalse(state.is_error())
        self.assertFalse(state.is_success())
        self.assertEqual(state.created_at, created_at)
        self.assertEqual(state.started_at, started_at)
        # Create an exception to get error state fromrunning state
        error = state.error(messages=['Error', 'State'])
        self.assertEqual(error.created_at, state.created_at)
        self.assertEqual(error.started_at, state.started_at)
        self.assertEqual(len(error.messages), 2)
        self.assertEqual(error.messages[0], 'Error')
        self.assertEqual(error.messages[1], 'State')

    def test_success_state(self):
        """Test creating instances of the success state class."""
        created_at = dt.datetime.now()
        started_at = created_at + dt.timedelta(seconds=10)
        finished_at = started_at + dt.timedelta(seconds=10)
        state = StateSuccess(
            created_at=created_at,
            started_at=started_at,
            finished_at=finished_at
        )
        self.assertTrue(state.is_success())
        self.assertFalse(state.is_error())
        self.assertFalse(state.is_pending())
        self.assertFalse(state.is_running())
        self.assertFalse(state.is_active())
        self.assertEqual(state.created_at, created_at)
        self.assertEqual(state.started_at, started_at)
        self.assertEqual(state.finished_at, finished_at)
        self.assertEqual(len(state.resources), 0)
        state = StateSuccess(
            created_at=created_at,
            started_at=started_at,
            finished_at=finished_at,
            resources=[FileResource('myfile', LOCAL_FILE)]
        )
        self.assertEqual(state.created_at, created_at)
        self.assertEqual(state.started_at, started_at)
        self.assertEqual(state.finished_at, finished_at)
        self.assertEqual(len(state.resources), 1)


if __name__ == '__main__':
    import unittest
    unittest.main()



# This file is part of the Reproducible Open Benchmarks for Data Analysis
# Platform (ROB).
#
# Copyright (C) 2019 NYU.
#
# ROB is free software; you can redistribute it and/or modify it under the
# terms of the MIT License; see LICENSE file for more details.

"""Implementation of the template loader for benchmark workflow templates. A
benchmark template has one additional element in its serialization that contains
the specification of the result schema.
"""

from benchtmpl.workflow.benchmark.base import BenchmarkTemplate
from benchtmpl.workflow.benchmark.schema import BenchmarkResultSchema
from benchtmpl.workflow.template.loader import DefaultTemplateLoader

import benchtmpl.error as err


"""Additional top-level elements of dictionary serialization for benchmark
template handles.
"""
LABEL_RESULTS = 'results'


class BenchmarkTemplateLoader(DefaultTemplateLoader):
    """Implementation of the template loader for benchmark workflow templates.
    """
    def from_dict(self, doc, identifier=None, base_dir=None, validate=True):
        """Create an instance of the benchmark template from a dictionary
        serialization. Expects a dictionary that contains the three top-level
        elements of the template handle plus the 'result' schema.

        Parameters
        ----------
        dict: dict
            Dictionary serialization of a workflow template
        identifier: string, optional
            Unique template identifier. This value will override the value in
            the document.
        base_dir: string, optional
            Optional path to directory on disk that contains static files that
            are required to run the represented workflow. This value will
            override the value in the document.
        validate: bool, optional
            Flag indicating if given template parameter declarations are to be
            validated against the parameter schema or not.

        Returns
        -------
        benchtmpl.workflow.benchmark.base.BenchmarkTemplate

        Raises
        ------
        benchtmpl.error.InvalidTemplateError
        benchtmpl.error.UnknownParameterError
        """
        # Ensure that the mandatory elements are present
        if not LABEL_RESULTS in doc:
            raise err.InvalidTemplateError('missing element \'{}\''.format(LABEL_RESULTS))
        # Get handle for workflow template from super class
        template = super(BenchmarkTemplateLoader, self).from_dict(
            doc=doc,
            identifier=identifier,
            base_dir=base_dir,
            validate=validate
        )
        # Get schema object from serialization
        try:
            schema = BenchmarkResultSchema.from_dict(doc[LABEL_RESULTS])
        except ValueError as ex:
            raise err.InvalidTemplateError(str(ex))
        return BenchmarkTemplate(
            identifier=template.identifier,
            base_dir=template.base_dir,
            workflow_spec=template.workflow_spec,
            parameters=template.parameters.values(),
            schema=schema
        )

    def to_dict(self, template):
        """Get dictionary serializationfor the wrokflow template.

        Parameters
        ----------
        template: benchtmpl.workflow.benchmark.base.BenchmarkTemplate
            Expects an instance of a benchmark template handle

        Returns
        -------
        dict
        """
        # Add serialization of schema to the serialization of the super class
        obj = super(BenchmarkTemplateLoader, self).to_dict(template)
        obj[LABEL_RESULTS] = template.schema.to_dict()
        return obj



# This file is part of the Reproducible Open Benchmarks for Data Analysis
# Platform (ROB).
#
# Copyright (C) 2019 NYU.
#
# ROB is free software; you can redistribute it and/or modify it under the
# terms of the MIT License; see LICENSE file for more details.

"""Simple file store that is used to maintain files that are uploaded by a user
(e.g., via an API). The file store assigns each uploaded file an unique
identifier while also maintaining the original file name.

When creating an instance for a parameterized workflow the user will pass the
identifier of an uploaded file as the argument for template parameters of the
respective data type for input files. These identifier will be replaces with
references to the actual files on disk managed by the file store.
"""

import os
import shutil

from benchtmpl.io.files.base import FileHandle
import benchtmpl.util.core as util


class Filestore(object):
    """The simple file store maintains a list of uploaded files. Each file is
    assigned a unique identifier that can later be used as argument value when
    creating a workflow instance for a workflow template.

    All files are maintained in sub-folders under a given base directory.
    """
    def __init__(self, directory):
        """Initialize the base directory where files are stored.

        Parameters
        ----------
        directory : string
            Path to the base directory.
        """
        self.directory = os.path.abspath(directory)
        # Create directory if it does not exist
        util.create_dir(self.directory)

    def delete_file(self, identifier):
        """Delete file with given identifier. Returns True if file was deleted
        or False if no such file existed.

        Parameters
        ----------
        identifier: string
            Unique file identifier

        Returns
        -------
        bool
        """
        file_dir =self.get_file_dir(identifier)
        if os.path.isdir(file_dir):
            shutil.rmtree(file_dir, ignore_errors=True)
            return True
        return False

    def get_file(self, identifier):
        """Get handle for file with given identifier. Returns None if no file
        with given identifier exists.

        Parameters
        ----------
        identifier: string
            Unique file identifier

        Returns
        -------
        benchtmpl.io.files.base.FileHandle
        """
        file_dir = self.get_file_dir(identifier)
        if os.path.isdir(file_dir):
            # The uploaded file is the only file in the directory
            file_name = os.listdir(file_dir)[0]
            return FileHandle(
                identifier=identifier,
                filepath=os.path.join(file_dir, file_name),
                file_name=file_name
            )
        return None

    def get_file_dir(self, identifier, create=False):
        """Get path to the directory for the file with the given identifier. If
        the directory does not exist it will be created if the create flag is
        True.

        Parameters
        ----------
        identifier: string
            Unique file identifier
        create: bool, optional
            File directory will be created if it does not exist and the flag
            is True

        Returns
        -------
        string
        """
        file_dir = os.path.join(os.path.abspath(self.directory), identifier)
        if create and not os.path.isdir(file_dir):
            os.makedirs(file_dir)
        return file_dir

    def list_files(self):
        """Get list of file handles for all uploaded files.

        Returns
        -------
        list(benchtmpl.io.files.base.FileHandle)
        """
        result = list()
        for f_name in os.listdir(self.directory):
            dir_name = os.path.join(self.directory, f_name)
            if os.path.isdir(dir_name):
                file_name = os.listdir(dir_name)[0]
                f_handle = FileHandle(
                    identifier=f_name,
                    filepath=os.path.join(dir_name, file_name),
                    file_name=file_name
                )
                result.append(f_handle)
        return result

    def upload_file(self, filename):
        """Create a new entry from a given local file. Will make a copy of the
        given file.

        Raises ValueError if the given file does not exist.

        Parameters
        ----------
        filename: string
            Path to file on disk

        Returns
        -------
        benchtmpl.io.files.base.FileHandle
        """
        # Ensure that the given file exists
        if not os.path.isfile(filename):
            raise ValueError('invalid file path \'' + str(filename) + '\'')
        file_name = os.path.basename(filename)
        # Create a new unique identifier for the file.
        identifier = util.get_unique_identifier()
        file_dir = self.get_file_dir(identifier, create=True)
        output_file = os.path.join(file_dir, file_name)
        # Copy the uploaded file
        shutil.copyfile(filename, output_file)
        # Add file to file index
        f_handle = FileHandle(
            identifier=identifier,
            filepath=output_file,
            file_name=file_name
        )
        return f_handle

    def upload_stream(self, file, file_name):
        """Create a new entry from a given file stream. Will copy the given
        file to a file in the base directory.

        Parameters
        ----------
        file: werkzeug.datastructures.FileStorage
            File object (e.g., uploaded via HTTP request)
        file_name: string
            Name of the file

        Returns
        -------
        benchtmpl.io.files.base.FileHandle
        """
        # Create a new unique identifier for the file.
        identifier = util.get_unique_identifier()
        file_dir = self.get_file_dir(identifier, create=True)
        output_file = os.path.join(file_dir, file_name)
        # Save the file object to the new file path
        file.save(output_file)
        f_handle = FileHandle(
            identifier=identifier,
            filepath=output_file,
            file_name=file_name
        )
        return f_handle



"""Test functionality of the default template repository implementation."""

import os
import shutil

from unittest import TestCase

from benchtmpl.workflow.template.base import TemplateHandle
from benchtmpl.workflow.template.loader import DefaultTemplateLoader
from benchtmpl.workflow.template.repo import TemplateRepository
from benchtmpl.workflow.template.repo import STATIC_FILES_DIR, TEMPLATE_FILE

import benchtmpl.error as err


ERR_SPEC = './tests/files/template-error-1.yaml'
JSON_SPEC = './tests/files/template/reana-template.json'
TMP_DIR = './tests/files/.tmp'
WORKFLOW_DIR = './tests/files/template'


class DummyIDFunc():
    """Dummy id function."""
    def __init__(self):
        self.count = 0

    def __call__(self):
        self.count += 1
        return '0000'


class TestTemplateRepository(TestCase):
    """Test functionality of default template repository."""
    def setUp(self):
        """Remove temporary template repository directory. The directory will
        be created when the repository is first instantiated.
        """
        if os.path.isdir(TMP_DIR):
            shutil.rmtree(TMP_DIR)

    def tearDown(self):
        """Remove temporary template repository directory."""
        if os.path.isdir(TMP_DIR):
            shutil.rmtree(TMP_DIR)

    def test_add_template(self):
        """Test creating templates."""
        store = TemplateRepository(base_dir=TMP_DIR)
        template = store.add_template(src_dir=WORKFLOW_DIR)
        self.validate_template_handle(template)
        # Ensure that the template handle has been serialized correctly
        f = os.path.join(store.base_dir, template.identifier, TEMPLATE_FILE)
        doc = DefaultTemplateLoader().load(f)
        d = os.path.join(store.base_dir, template.identifier, STATIC_FILES_DIR)
        self.assertTrue(os.path.isdir(d))
        # Get template and repeat tests
        self.validate_template_handle(store.get_template(template.identifier))
        store = TemplateRepository(base_dir=TMP_DIR)
        self.validate_template_handle(store.get_template(template.identifier))
        # Add template with JSON specification file
        template = store.add_template(
            src_dir=WORKFLOW_DIR,
            template_spec_file=JSON_SPEC
        )
        self.validate_template_handle(template)
        # Unknown template error
        with self.assertRaises(err.UnknownTemplateError):
            store.get_template('unknown')
        # Errors when specifying wrong parameter combination
        with self.assertRaises(ValueError):
            store.add_template()
        with self.assertRaises(ValueError):
            store.add_template(src_dir=WORKFLOW_DIR, src_repo_url=WORKFLOW_DIR)
        # Load templates with erroneous specifications
        with self.assertRaises(err.InvalidTemplateError):
            store.add_template(src_dir=WORKFLOW_DIR, template_spec_file=ERR_SPEC)
        # Error when cloning invalid repository from GitHub
        with self.assertRaises(err.InvalidTemplateError):
            store.add_template(src_repo_url='https://github.com/reanahub/reana-demo-helloworld')

    def test_delete_template(self):
        """Ensure correct return values when deleting existing and non-existing
        templates.
        """
        store = TemplateRepository(base_dir=TMP_DIR)
        template = store.add_template(src_dir=WORKFLOW_DIR)
        f = os.path.join(store.base_dir, template.identifier, TEMPLATE_FILE)
        d = os.path.join(store.base_dir, template.identifier, STATIC_FILES_DIR)
        self.assertTrue(os.path.isfile(f))
        self.assertTrue(os.path.isdir(d))
        self.assertTrue(store.delete_template(template.identifier))
        self.assertFalse(os.path.isfile(f))
        self.assertFalse(os.path.isdir(d))
        self.assertFalse(store.delete_template(template.identifier))
        # Test deleting after store object is re-instantiated
        template = store.add_template(src_dir=WORKFLOW_DIR)
        store = TemplateRepository(base_dir=TMP_DIR)
        self.assertTrue(store.delete_template(template.identifier))
        self.assertFalse(store.delete_template(template.identifier))

    def test_error_for_id_func(self):
        """Error when the id function cannot return unique folder identifier."""
        dummy_func = DummyIDFunc()
        store = TemplateRepository(base_dir=os.path.join(TMP_DIR), id_func=dummy_func)
        os.makedirs(os.path.join(store.base_dir, dummy_func()))
        with self.assertRaises(RuntimeError):
            store.add_template(src_dir=WORKFLOW_DIR)
        self.assertEqual(dummy_func.count, 102)

    def validate_template_handle(self, template):
        """Basic tests to validate a given template handle."""
        self.assertIsNotNone(template.identifier)
        self.assertIsNotNone(template.base_dir)
        self.assertEqual(
            template.workflow_spec['inputs']['files'],
            ['$[[code]]', '$[[names]]']
        )
        self.assertEqual(len(template.parameters), 4)
        self.assertTrue(os.path.isfile(os.path.join(template.base_dir, 'code/helloworld.py')))
        self.assertTrue(os.path.isfile(os.path.join(template.base_dir, 'inputs/names.txt')))


if __name__ == '__main__':
    import unittest
    unittest.main()



# This file is part of the Reproducible Open Benchmarks for Data Analysis
# Platform (ROB).
#
# Copyright (C) 2019 NYU.
#
# ROB is free software; you can redistribute it and/or modify it under the
# terms of the MIT License; see LICENSE file for more details.

"""Helper methods for the reproducible benchmark engine. Provides methods to
read and write files in JSON and YAML format and for the creation of unique
identifiers.
"""

import errno
import json
import os
import uuid
import yaml


"""Identifier for supported data formats."""
FORMAT_JSON = 'JSON'
FORMAT_YAML = 'YAML'


def create_dir(directory):
    """Safely create the given directory path if it does not exist.

    Parameters
    ----------
    directory: string
        Path to directory that is being created.
    """
    # Based on https://stackoverflow.com/questions/273192/how-can-i-safely-create-a-nested-directory
    if not os.path.exists(directory):
        try:
            os.makedirs(directory)
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise


def get_unique_identifier():
    """Create a new unique identifier.

    Returns
    -------
    string
    """
    return str(uuid.uuid4()).replace('-', '')


def get_short_identifier():
    """Create a unique identifier that contains only eigth characters. Uses the
    prefix of a unique identifier as the result.

    Returns
    -------
    string
    """
    return get_unique_identifier()[:8]


def read_object(filename, format=None):
    """Load a Json object from a file. The file may either be in Yaml or in Json
    format.

    Parameters
    ----------
    filename: string
        Path to file on disk
    format: string, optional
        Optional file format identifier. The default is YAML

    Returns
    -------
    dict

    Raises
    ------
    ValueError
    """
    # Guess format based on file suffix if not given
    if format is None:
        if filename.endswith('.json'):
            format = FORMAT_JSON
        else:
            format = FORMAT_YAML
    if format.upper() == FORMAT_YAML:
        with open(filename, 'r') as f:
            try:
                return yaml.load(f.read(), Loader=yaml.FullLoader)
            except yaml.parser.ParserError as ex:
                raise ValueError(ex)
    elif format.upper() == FORMAT_JSON:
        with open(filename, 'r') as f:
            return json.load(f)
    else:
        raise ValueError('unknown data format \'' + str(format) + '\'')


def write_object(filename, obj, format=None):
    """Write given dictionary to file as Json object.

    Parameters
    ----------
    filename: string
        Path to output file
    obj: dict
        Output object

    Raises
    ------
    ValueError
    """
    if format is None:
        if filename.endswith('.json'):
            format = FORMAT_JSON
        else:
            format = FORMAT_YAML
    if format.upper() == FORMAT_YAML:
        with open(filename, 'w') as f:
            yaml.dump(obj, f)
    elif format.upper() == FORMAT_JSON:
        with open(filename, 'w') as f:
            json.dump(obj, f)
    else:
        raise ValueError('unknown data format \'' + str(format) + '\'')



# This file is part of the Reproducible Open Benchmarks for Data Analysis
# Platform (ROB).
#
# Copyright (C) 2019 NYU.
#
# ROB is free software; you can redistribute it and/or modify it under the
# terms of the MIT License; see LICENSE file for more details.

"""Definition of schema components for benchmark result schemas. The schema
definition is part of the extended workflow template specification that is used
to define benchmarks.
"""

import benchtmpl.workflow.parameter.declaration as pd

"""Supported data types for result values."""
DATA_TYPES = [pd.DT_DECIMAL, pd.DT_INTEGER, pd.DT_STRING]


"""Labels for serialization."""
LABEL_ID = 'id'
LABEL_IS_DEFAULT = 'isDefault'
LABEL_NAME = 'name'
LABEL_RESULT_FILE = 'file'
LABEL_REQUIRED = 'required'
LABEL_SCHEMA = 'schema'
LABEL_SORT_ORDER = 'sortOrder'
LABEL_TYPE = 'type'

COLUMN_LABELS = [LABEL_ID, LABEL_NAME, LABEL_TYPE]
SCHEMA_LABELS = [LABEL_RESULT_FILE, LABEL_SCHEMA]


"""Column sort orders."""
SORT_ASC = 'ASC'
SORT_DESC = 'DESC'

class BenchmarkResultColumn(object):
    """Column in the result schema of a benchmark. Each column has a unique
    identifier and unique name. The identifier is used as column name in the
    database schema. the name is for display purposes in the user interface.
    """
    def __init__(self, identifier, name, data_type, required=None, is_default=None, sort_order=None):
        """Initialize the unique column identifier, name, and the data type. If
        the value of data_type is not in the list of supported data types an
        error is raised.

        Parameters
        ----------
        identifier: string
            Unique column identifier
        name: string
            Unique column name
        data_type: string
            Data type identifier
        required: bool, optional
            Indicates whether a value is expected for this column in every
            benchmark run result
        is_default: bool, optional
            Indicates whether this column is the default csort column for the
            benchmark leaderboard
        sort_order: string, optional
            Sort order for this column when generating the leaderboard
        """
        # Raise error if the data type value is not in the list of supported
        # data types
        if not data_type in DATA_TYPES:
            raise ValueError('unknown data type \'{}\''.format(data_type))
        self.identifier = identifier
        self.name = name
        self.data_type = data_type
        self.required = required if not required is None else True
        self.is_default = is_default if not is_default is None else False
        self.sort_order = sort_order if not sort_order is None else SORT_DESC

    @staticmethod
    def from_dict(doc):
        """Get an instance of the column from the dictionary serialization.
        Raises an error if the given dictionary does not contain the expected
        elements as generated by the to_dict() method of the class.

        Parameters
        ----------
        doc: dict
            Dictionary serialization of a column object

        Returns
        -------
        benchtmpl.workflow.benchmark.schema.BenchmarkResultColumn

        Raises
        ------
        ValueError
        """
        # Validate the serialization dictionary
        validate_doc(
            doc,
            COLUMN_LABELS,
            optional_labels=[LABEL_REQUIRED, LABEL_IS_DEFAULT, LABEL_SORT_ORDER]
        )
        # Return instance of the column object
        return BenchmarkResultColumn(
            identifier=doc[LABEL_ID],
            name=doc[LABEL_NAME],
            data_type=doc[LABEL_TYPE],
            required=doc.get(LABEL_REQUIRED),
            is_default=doc.get(LABEL_IS_DEFAULT),
            sort_order=doc.get(LABEL_SORT_ORDER)
        )

    def is_desc(self):
        """True if the values in this columns are sorted in descending order
        when generating the leaderboard.

        Returns
        -------
        bool
        """
        return self.sort_order == SORT_DESC

    def sort_statement(self):
        """Get column identifier and the optional sort order statement depending
        on the current value for the sort order attribute. The result is to be
        used when generating SQL queries that fetch benchmark results.

        Returns
        -------
        string
        """
        if self.is_desc():
            return self.identifier + ' DESC'
        else:
            return self.identifier

    def sql_stmt(self):
        """SQL statement for column in create table statement.

        Returns
        -------
        string
        """
        stmt = self.identifier
        if self.data_type == pd.DT_INTEGER:
            stmt += ' INTEGER'
        elif self.data_type == pd.DT_DECIMAL:
            stmt += ' DOUBLE'
        else:
            stmt += ' TEXT'
        if self.required:
            stmt += ' NOT NULL'
        return stmt

    def to_dict(self):
        """Get dictionary serialization for the column object.

        Returns
        -------
        dict
        """
        return {
            LABEL_ID: self.identifier,
            LABEL_NAME: self.name,
            LABEL_TYPE: self.data_type,
            LABEL_REQUIRED: self.required,
            LABEL_IS_DEFAULT: self.is_default,
            LABEL_SORT_ORDER: self.sort_order
        }


class BenchmarkResultSchema(object):
    """The result schema of a benchmark run is a collection of columns. The
    result object of the run is expected to contain a value for each required
    column and at most for all columns. The schema also contains the identifier
    of the output file that contains the result object.
    """
    def __init__(self, columns, result_file_id):
        """Initialize the schema columns and the result file identifier.

        Parameters
        ----------
        columns: list(benchtmpl.workflow.benchmark.schema.BenchmarkResultColumn)
            List of columns in the result object
        result_file_id: string
            Identifier of the benchmark run result file that contains the
            analytics results.
        """
        self.columns = columns
        self.result_file_id = result_file_id

    @staticmethod
    def from_dict(doc):
        """Get an instance of the schema from a dictionary serialization.
        Raises an error if the given dictionary does not contain the expected
        elements as generated by the to_dict() method of the class or if the
        names or identifier of columns are not unique.

        Parameters
        ----------
        doc: dict
            Dictionary serialization of a benchmark result schema object

        Returns
        -------
        benchtmpl.workflow.benchmark.schema.BenchmarkResultSchema

        Raises
        ------
        ValueError
        """
        # Validate the serialization dictionary
        validate_doc(doc, SCHEMA_LABELS)
        # Get column list. Ensure that all column names and identifier are
        # unique
        columns=[BenchmarkResultColumn.from_dict(c) for c in doc[LABEL_SCHEMA]]
        ids = set()
        names = set()
        for col in columns:
            if col.identifier in ids:
                raise ValueError('duplicate column identifier \'{}\''.format(col.identifier))
            ids.add(col.identifier)
            if col.name in names:
                raise ValueError('not unique column name \'{}\''.format(col.name))
            names.add(col.name)
        file_id = doc[LABEL_RESULT_FILE]
        # Return instance of the column object
        return BenchmarkResultSchema(columns=columns, result_file_id=file_id)

    def to_dict(self):
        """Get dictionary serialization for the result schema object.

        Returns
        -------
        dict
        """
        return {
            LABEL_RESULT_FILE: self.result_file_id,
            LABEL_SCHEMA: [col.to_dict() for col in self.columns]
        }


# ------------------------------------------------------------------------------
# Helper Methods
# ------------------------------------------------------------------------------

def validate_doc(doc, mandatory_labels, optional_labels=[]):
    """Raises error if the dictionary contains labels that are not in the given
    label lists or if there are labels in the mandatory list that are not in the
    dictionary.

    Paramaters
    ----------
    doc: dict
        Dictionary serialization of an object
    mandatory_labels: list(string)
        List of mandatory labels for the dictionary serialization
    optional_labels: list(string), optional
        List of optional labels for the dictionary serialization

    Raises
    ------
    ValueError
    """
    # Ensure that all mandatory labels are present in the dictionary
    for key in mandatory_labels:
        if not key in doc:
            raise ValueError('missing element \'{}\''.format(key))
    # Raise error if additional elements are present in the dictionary
    labels = mandatory_labels + optional_labels
    for key in doc:
        if not key in labels:
            raise ValueError('unknown element \'{}\''.format(key))



"""Utilities for workflow templates. Provides methods and classes to copy
and upload files for workflow runs.
"""

import errno
import os
import shutil

import benchtmpl.error as err
import benchtmpl.workflow.template.base as tmpl



# ------------------------------------------------------------------------------
# Helper Classes
# ------------------------------------------------------------------------------

class FileCopy(object):
    """File upload function that copies files to a local directory instead of
    uploading them to a remote server. This class is used as argument for the
    REANA upload function that prepares a workflow run.
    """
    def __init__(self, destination_dir):
        """Initialize the destination directory into which all files are copied.

        Parameters
        ----------
        destination_dir: string
            Path to local directory
        """
        self.destination_dir = destination_dir

    def __call__(self, source, target):
        """Copy the local source file to the relative target location in the
        destination directory.

        Parameters
        ----------
        source: string
            Path to file on disk
        target: string
            Relative target path for file in destination directory
        """
        dst = os.path.join(self.destination_dir, target)
        # If the source references a directory the whole directory tree is
        # copied
        if os.path.isdir(source):
            shutil.copytree(src=source, dst=dst)
        else:
            # Based on https://stackoverflow.com/questions/2793789/create-destination-path-for-shutil-copy-files/3284204
            try:
                shutil.copy(src=source, dst=dst)
            except IOError as e:
                # ENOENT(2): file does not exist, raised also on missing dest
                # parent dir
                if e.errno != errno.ENOENT or not os.path.isfile(source):
                    raise
                # try creating parent directories
                os.makedirs(os.path.dirname(dst))
                shutil.copy(src=source, dst=dst)


# ------------------------------------------------------------------------------
# Helper Methods
# ------------------------------------------------------------------------------

def upload_files(template, files, arguments, loader):
    """Upload all references to local files in a given list of file names of
    parameter references. THe list of files, for example corresponds to the
    entries in the 'inputs.files' section of a REANA workflow specification.

    Uses a loader function to allow use of this method in cases where the
    workflow is executed locally or remote using a REANA cluster instance.

    Raises errors if (i) an unknown parameter is referenced or (ii) if the type
    of a referenced parameter in the input files section is not of type file.

    Parameters
    ----------
    template: benchtmpl.workflow.template.base.TemplateHandle
        Workflow template containing the parameterized specification and the
        parameter declarations
    files: list(string)
        List of file references
    arguments: dict(benchtmpl.workflow.parameter.value.TemplateArgument)
        Dictionary of argument values for parameters in the template
    loader: func
        File (up)load function that takes a filepath as the first argument and
        a (remote) target path as the second argument

    Raises
    ------
    benchtmpl.error.InvalidTemplateError
    benchtmpl.error.MissingArgumentError
    benchtmpl.error.UnknownParameterError
    """
    for val in files:
        # Set source and target values depending on whether the list
        # entry references a template parameter or not
        if tmpl.is_parameter(val):
            var = tmpl.get_parameter_name(val)
            # Raise error if the type of the referenced parameter is
            # not file
            para = template.get_parameter(var)
            if not para.is_file():
                raise err.InvalidTemplateError('expected file parameter for \'{}\''.format(var))
            arg = arguments.get(var)
            if arg is None:
                if para.default_value is None:
                    raise err.MissingArgumentError(var)
                else:
                    # Set source path to default value (assuming that
                    # the default points to a file in the template
                    # base directory)
                    source = os.path.join(
                        template.base_dir,
                        para.default_value
                    )
            else:
                # Get path to source file from file handle
                source = arg.value.filepath
            if para.has_constant():
                target = para.get_constant()
            else:
                target = arg.value.name
        else:
            source = os.path.join(template.base_dir, val)
            target = val
        # Upload source file
        loader(source, target)



# This file is part of the Reproducible Open Benchmarks for Data Analysis
# Platform (ROB).
#
# Copyright (C) 2019 NYU.
#
# ROB is free software; you can redistribute it and/or modify it under the
# terms of the MIT License; see LICENSE file for more details.

"""Specification for workflow template parameter declarations. This module
defines the Json schema for parameter declarations that are part of a workflow
template.

In addition, helper methods are provided for convenience to create parameter
declarations from within Python scripts.
"""

from jsonschema import validate, ValidationError

from benchtmpl.error import InvalidParameterError


# ------------------------------------------------------------------------------
# Schema
# ------------------------------------------------------------------------------

"""Labels for elements in the schema of a parameter declaration."""
LABEL_AS = 'as'
LABEL_DATATYPE = 'datatype'
LABEL_DEFAULT = 'defaultValue'
LABEL_DESCRIPTION = 'description'
LABEL_ID = 'id'
LABEL_IS_DEFAULT = 'isDefault'
LABEL_INDEX = 'index'
LABEL_NAME = 'name'
LABEL_PARENT = 'parent'
LABEL_REQUIRED = 'required'
LABEL_VALUE = 'value'
LABEL_VALUES = 'values'


PARAMETER_SCHEMA = {
    'type': 'object',
    'properties': {
        LABEL_ID: {'type': 'string'},
        LABEL_NAME: {'type': 'string'},
        LABEL_DESCRIPTION: {'type': 'string'},
        LABEL_DATATYPE: {'type': 'string'},
        LABEL_PARENT: {'type': 'string'},
        LABEL_VALUES: {
            'type': 'array',
            'items': {
                'type': 'object',
                'properties': {
                    LABEL_IS_DEFAULT: {'type': 'boolean'},
                    LABEL_NAME: {'type': 'string'},
                    LABEL_VALUE: {'oneOf': [
                        {'type': 'boolean'},
                        {'type': 'string'},
                        {'type': 'number'}
                    ]}
                },
                'required': [LABEL_VALUE]
            }
        },
        LABEL_REQUIRED: {'type': 'boolean'},
        LABEL_DEFAULT: {'oneOf': [
            {'type': 'boolean'},
            {'type': 'string'},
            {'type': 'number'}
        ]},
        LABEL_INDEX: {'type': 'number'},
        LABEL_AS: {'type': 'string'}
    },
    'required': [LABEL_ID]
}


# ------------------------------------------------------------------------------
# Data types for template parameters
# ------------------------------------------------------------------------------

"""Definition of parameter data types."""
DT_BOOL = 'bool'
DT_DECIMAL = 'decimal'
DT_FILE = 'file'
DT_INTEGER = 'int'
DT_LIST = 'list'
DT_RECORD = 'record'
DT_STRING = 'string'

DATA_TYPES = [
    DT_BOOL,
    DT_DECIMAL,
    DT_FILE,
    DT_INTEGER,
    DT_LIST,
    DT_RECORD,
    DT_STRING
]


# ------------------------------------------------------------------------------
# Helper Methods
# ------------------------------------------------------------------------------

def enum_value(value, text=None, is_default=False):
    """Create dictionary representing a value in an enumeration of possible
    values for a parameter.

    Parameters
    ----------
    value: int, float or string
        Enumeration value
    text: string, optional
        Text representation for the value in an front-end form
    is_default: bool, optional
        Flag indicating whether this is the default value for the list

    Returns
    -------
    dict
    """
    obj = {LABEL_VALUE: value, LABEL_IS_DEFAULT: is_default}
    if not text is None:
        obj[LABEL_NAME] = text
    else:
        obj[LABEL_NAME] = str(value)
    return obj


def parameter_declaration(
        identifier, name=None, data_type=DT_STRING, description=None, index=0,
        required=True, values=None, parent=None, default_value=None,
        as_const=None
    ):
    """Create a dictionary that contains a module parameter specification.

    Raises InvalidParameterError if an invalid data type is given or if the
    identifier is None.

    Parameters
    ----------
    identifier: string
        Unique parameter identifier
    name: string, optional
        Printable parameter name. The default value is the parameter identifier.
    data_type: string, optional
        Parameter type. The default value is DT_STRING
    description: string, optional
        Optional text providing a more comprehensive description for the
        parameter. Default is the parameter name.
    index: int, optional
        Index position of argument in input form. The default value is 0.
    required: bool, optional
        Flag indicating whether values for this parameter are required or
        optional
    values: list, optional
        List of valid parameter values (for selection in a front-end form)
    parent: string, optional
        Identifier of a grouping element
    default_value: bool, string, number, optional
        Optional default value for a scalar parameter. Default values for file
        parameters are ignored.
    as_const: string, None
        Constant replacement value. This is primarily used to replace the name
        of uploaded files with a constant value.

    Returns
    -------
    dict

    Raises
    ------
    benchtmpl.error.InvalidParameterError
    """
    if identifier is None:
        raise InvalidParameterError('missing identifier')
    if not data_type in DATA_TYPES:
        raise InvalidParameterError('invalid parameter data type \'{}\''.format(data_type))
    para = {
        LABEL_ID: identifier,
        LABEL_NAME: name if not name is None else identifier,
        LABEL_DATATYPE: data_type,
        LABEL_INDEX: index,
        LABEL_REQUIRED: required
    }
    # Set optional properties
    if not description is None:
        para[LABEL_DESCRIPTION] = description
    else:
        para[LABEL_DESCRIPTION] = para[LABEL_NAME]
    if not values is None:
        para[LABEL_VALUES] = values
    if not parent is None:
        para[LABEL_PARENT] = parent
    if not as_const is None:
        para[LABEL_AS] = as_const
    # Ignore given default values for file parameters
    if not default_value is None and data_type != DT_FILE:
        para[LABEL_DEFAULT] = default_value
    return para


def set_defaults(obj):
    """Set default values for parameter declaration if the respective element
    is missing in the given object. It is assumed that the parameter identifier
    is set in the given object.

    Returns a (modified) copy of the given parameter declaration object.

    Parameters
    ----------
    obj: dict
        Dictionary containing a parameter declaration

    Returns
    -------
    dict
    """
    para = dict(obj)
    # Set name to identifier value if no name is present
    set_value(para, LABEL_NAME, obj[LABEL_ID])
    # Set data type to DT_STRING if no data type is present
    set_value(para, LABEL_DATATYPE, DT_STRING)
    # Set description to name value if no description is present
    set_value(para, LABEL_DESCRIPTION, para[LABEL_NAME])
    # Set index to 0 if no parameter index is present
    set_value(para, LABEL_INDEX, 0)
    # Set required flag to True if not present
    set_value(para, LABEL_REQUIRED, True)
    return para


def set_value(parameter, key, value):
    """Set the value for the parameter element with given key to the given value
    if the element is currently not present in the parameter dictionary. If the
    element with key is present no changes occur.

    Parameters
    ----------
    parameter: dict
        Dictionary containing a parameter declaration
    key: string
        Parameter element label
    value: any
        Default element value
    """
    if not key in parameter:
        parameter[key] = value


def validate_parameter(param_declaration):
    """Validate a given parameter declaration.

    Raises a InvalidParameterError if an invalid parameter declaration is given
    or if the data type for the parameter is invalid.

    Parameters
    ----------
    param_declaration: dict
        Dictionary containing parameter declaration

    Raises
    ------
    benchtmpl.error.InvalidParameterError
    """
    # Make sure that the given package declaration matches the schema
    try:
        validate(param_declaration, PARAMETER_SCHEMA)
    except ValidationError as ex:
        raise InvalidParameterError('failed to validate parameter declaration. {}'.format(ex.message))
    # Ensure that the given parameter data type is valid
    dt = param_declaration[LABEL_DATATYPE]
    if not dt in DATA_TYPES:
        raise InvalidParameterError('invalid data type \'{}\''.format(dt))



"""Test TemplateHandle functionality."""

from unittest import TestCase

from benchtmpl.io.files.base import FileHandle
from benchtmpl.workflow.parameter.base import TemplateParameter
from benchtmpl.workflow.resource.base import ResourceDescriptor, LABEL_ID
from benchtmpl.workflow.template.base import TemplateHandle
from benchtmpl.workflow.template.loader import DefaultTemplateLoader

import benchtmpl.error as err
import benchtmpl.workflow.parameter.declaration as pd
import benchtmpl.workflow.template.base as tmpl
import benchtmpl.workflow.template.loader as loader


TEMPLATE_JSON_FILE = 'tests/files/template/template.json'
TEMPLATE_YAML_FILE = 'tests/files/template/template.yaml'
TEMPLATE_ERR = './tests/files/template-error-2.yaml'


class TestTemplateHandle(TestCase):
    def test_duplicate_id(self):
        """Ensure that exception is raised if parameter identifier are not
        unique.
        """
        with self.assertRaises(err.InvalidTemplateError):
            DefaultTemplateLoader().from_dict({
                    loader.LABEL_WORKFLOW: dict(),
                    loader.LABEL_PARAMETERS: [
                        pd.parameter_declaration('A', index=1),
                        pd.parameter_declaration('B'),
                        pd.parameter_declaration('C'),
                        pd.parameter_declaration('A', index=2),
                        pd.parameter_declaration('E', index=1)
                    ]
                },
                validate=True
            )

    def test_get_parameter_references(self):
        """Test function to get all parameter references in a workflow
        specification."""
        spec = {
            'input': [
                'A',
                '$[[X]]',
                {'B': {'C': '$[[Y]]', 'D': [123, '$[[Z]]']}}
            ],
            'E': {'E': 'XYZ', 'F': 23, 'G': '$[[W]]'},
            'F': '$[[U]]',
            'G': ['$[[V]]', 123]
        }
        refs = tmpl.get_parameter_references(spec)
        self.assertEqual(refs, set(['U', 'V', 'W', 'X', 'Y', 'Z']))
        # If given parameter set as argument the elements in that set are part
        # of the result
        para = set(['A', 'B', 'X'])
        refs = tmpl.get_parameter_references(spec, parameters=para)
        self.assertEqual(refs, set(['A', 'B', 'U', 'V', 'W', 'X', 'Y', 'Z']))
        # Error if specification contains nested lists
        with self.assertRaises(err.InvalidTemplateError):
            tmpl.get_parameter_references({
                'input': [
                    'A',
                    ['$[[X]]'],
                    {'B': {'C': '$[[Y]]', 'D': [123, '$[[Z]]']}}
                ]
            })
        # Error when loading specification that references undefined parameter
        with self.assertRaises(err.UnknownParameterError):
            template = DefaultTemplateLoader().load(TEMPLATE_ERR, validate=True)

    def test_init(self):
        """Test initialization of attribuets and error cases when creating
        template instances.
        """
        th = TemplateHandle(
            workflow_spec=dict(),
            parameters=[
                TemplateParameter(pd.parameter_declaration('A')),
                TemplateParameter(pd.parameter_declaration('B'))
            ]
        )
        self.assertIsNotNone(th.identifier)
        self.assertIsNone(th.base_dir)
        th = TemplateHandle(
            identifier='ABC',
            base_dir='XYZ',
            workflow_spec=dict(),
            parameters=[
                TemplateParameter(pd.parameter_declaration('A')),
                TemplateParameter(pd.parameter_declaration('B'))
            ]
        )
        self.assertEqual(th.identifier, 'ABC')
        self.assertEqual(th.base_dir, 'XYZ')
        with self.assertRaises(err.InvalidTemplateError):
            TemplateHandle(
                workflow_spec=dict(),
                parameters=[
                    TemplateParameter(pd.parameter_declaration('A')),
                    TemplateParameter(pd.parameter_declaration('B')),
                    TemplateParameter(pd.parameter_declaration('A'))
                ]
            )

    def test_nested_parameters(self):
        """Test proper nesting of parameters for DT_LIST and DT_RECORD."""
        # Create a new TemplateHandle with an empty workflow specification and
        # a list of six parameters (one record and one list)
        template = DefaultTemplateLoader().from_dict({
                loader.LABEL_WORKFLOW: dict(),
                loader.LABEL_PARAMETERS: [
                    pd.parameter_declaration('A'),
                    pd.parameter_declaration('B', data_type=pd.DT_RECORD),
                    pd.parameter_declaration('C', parent='B'),
                    pd.parameter_declaration('D', parent='B'),
                    pd.parameter_declaration('E', data_type=pd.DT_LIST),
                    pd.parameter_declaration('F', parent='E'),
                ]
            },
            validate=True
        )
        # Parameters 'A', 'C', 'D', and 'F' have no children
        for key in ['A', 'C', 'D', 'F']:
            self.assertFalse(template.get_parameter(key).has_children())
        # Parameter 'B' has two children 'C' and 'D'
        b = template.get_parameter('B')
        self.assertTrue(b.has_children())
        self.assertEqual(len(b.children), 2)
        self.assertTrue('C' in [p.identifier for p in b.children])
        self.assertTrue('D' in [p.identifier for p in b.children])
        # Parameter 'E' has one childr 'F'
        e = template.get_parameter('E')
        self.assertTrue(e.has_children())
        self.assertEqual(len(e.children), 1)
        self.assertTrue('F' in [p.identifier for p in e.children])

    def test_serialization(self):
        """Test serialization of workflow templates."""
        template = TemplateHandle(
            identifier='ABC',
            base_dir='XYZ',
            workflow_spec=dict(),
            parameters=[
                TemplateParameter(pd.parameter_declaration('A')),
                TemplateParameter(pd.parameter_declaration('B', data_type=pd.DT_LIST)),
                TemplateParameter(pd.parameter_declaration('C', parent='B'))
            ]
        )
        doc = DefaultTemplateLoader().to_dict(template)
        parameters = DefaultTemplateLoader().from_dict(doc).parameters
        self.assertTrue(len(parameters), 2)
        self.assertTrue('A' in parameters)
        self.assertTrue('B' in parameters)
        self.assertEqual(len(parameters['B'].children), 1)
        template = DefaultTemplateLoader().from_dict(doc)
        self.assertEqual(template.identifier, 'ABC')
        # The base directory is not materialized
        self.assertIsNone(template.base_dir)
        # Invalid resource descriptor serializations
        with self.assertRaises(err.InvalidTemplateError):
            ResourceDescriptor.from_dict(dict())
        with self.assertRaises(err.InvalidTemplateError):
            ResourceDescriptor.from_dict({LABEL_ID: 'A', 'noname': 'B'})

    def test_simple_replace(self):
        """Replace parameter references in simple template with argument values.
        """
        for filename in [TEMPLATE_YAML_FILE, TEMPLATE_JSON_FILE]:
            template = DefaultTemplateLoader().load(filename)
            arguments = {
                'code': template.get_argument(
                    identifier='code',
                    value=FileHandle('code/helloworld.py')
                ),
                'names': template.get_argument(
                    identifier='names',
                    value=FileHandle('data/list-of-names.txt')
                ),
                'sleeptime': template.get_argument(
                    identifier='sleeptime',
                    value=10
                )
            }
            spec = tmpl.replace_args(
                spec=template.workflow_spec,
                arguments=arguments,
                parameters=template.parameters
            )
            self.assertEqual(spec['inputs']['files'][0], 'helloworld.py')
            self.assertEqual(spec['inputs']['files'][1], 'data/names.txt')
            self.assertEqual(spec['inputs']['parameters']['helloworld'], 'code/helloworld.py')
            self.assertEqual(spec['inputs']['parameters']['inputfile'], 'data/names.txt')
            self.assertEqual(spec['inputs']['parameters']['sleeptime'], 10)
            self.assertEqual(spec['inputs']['parameters']['waittime'], 5)

    def test_sort(self):
        """Test the sort functionality of the template list_parameters method.
        """
        # Create a new TemplateHandle with an empty workflow specification and
        # a list of five parameters
        template = DefaultTemplateLoader().from_dict({
                loader.LABEL_WORKFLOW: dict(),
                loader.LABEL_PARAMETERS: [
                    pd.parameter_declaration('A', index=1),
                    pd.parameter_declaration('B'),
                    pd.parameter_declaration('C'),
                    pd.parameter_declaration('D', index=2),
                    pd.parameter_declaration('E', index=1)
                ]
            },
            validate=True
        )
        # Get list of sorted parameter identifier from listing
        keys = [p.identifier for p in template.list_parameters()]
        self.assertEqual(keys, ['B', 'C', 'A', 'E', 'D'])


if __name__ == '__main__':
    import unittest
    unittest.main()



# This file is part of the Reproducible Open Benchmarks for Data Analysis
# Platform (ROB).
#
# Copyright (C) 2019 NYU.
#
# ROB is free software; you can redistribute it and/or modify it under the
# terms of the MIT License; see LICENSE file for more details.

"""Handles and descriptors for resources that are created by successful workflow
runs.

The resource descriptor is used to define the type of resources that are
generated by workflow runs.

Resource handles represent resources that are created as the result of a
successful workflow run. The most common type of resource at this point are
plain files.

In the future we might want to maintain additional information about the files
(e.g. serialization format and schema), as well as allow additional types of
resources.
"""

from benchtmpl.error import InvalidTemplateError


"""Identifier for resource types."""
RESOURCE_FILE = 'file'


"""Element labels for serialization of resource descriptors."""
LABEL_ID = 'id'
LABEL_TYPE = 'type'


class ResourceDescriptor(object):
    """Each resource handle has an identifier that is unique among the resources
    that are created by a workflow run. The identifier should remain the same
    over different runs for the same type of resource (e.g., an output file that
    is expected to be generated by a workflow run).

    The resource descriptor also has a type identifier.
    """
    def __init__(self, identifier, type_id):
        """Initialize the resource identifier and the type identifier.

        Parameters
        ----------
        identifier: string
            Resource identifier that is unique among all resources that are
            created within a single workflow run.
        type_id: string
            Identifier for resource type
        """
        self.identifier = identifier
        self.type_id = type_id

    @staticmethod
    def from_dict(doc):
        """Get instance of the resource descriptor from a dictionary
        serialization as created by the to_dict() method. Expects a dictionary
        with two elements 'id' and 'type'.

        Raises InvalidTemplateError if an invalid dictionary is given.

        Parameters
        ----------
        doc: dict
            Dictionary serialization of a resource descriptor

        Returns
        -------
        benchtmpl.workflow.resource.base.ResourceDescriptor

        Raises
        ------
        ValueError
        """
        # Expect exactly two elements in the dictionary
        if len(doc) != 2:
            raise InvalidTemplateError('invalid resource descriptor serialization')
        if not (LABEL_ID in doc and LABEL_TYPE in doc):
            raise InvalidTemplateError('invalid resource descriptor serialization')
        return ResourceDescriptor(
            identifier=doc[LABEL_ID],
            type_id=doc[LABEL_TYPE]
        )

    def to_dict(self):
        """Create dictionary serialization of the resource descriptor

        Returns
        -------
        dict
        """
        return {LABEL_ID: self.identifier, LABEL_TYPE: self.type_id}


class FileResource(ResourceDescriptor):
    """Handle for a file resource that is created as the result of a workflow
    run. The handle contains the reference to the persistent file that is stored
    on disk. The file should be maintaioned by the workflow backend in a
    persistent manner in order to be accessible as long as information about
    the workflow run is maintained in some backend database.
    """
    def __init__(self, identifier, filepath):
        """Initialize the resource identifier and the path to the (persistently)
        created file.

        Parameters
        ----------
        identifier: string
            Resource identifier that is unique among all resources that are
            created within a single workflow run.
        filepath: string
            Path to persistent file on disk.
        """
        super(FileResource, self).__init__(
            identifier=identifier,
            type_id=RESOURCE_FILE
        )
        self.filepath = filepath



# This file is part of the Reproducible Open Benchmarks for Data Analysis
# Platform (ROB).
#
# Copyright (C) 2019 NYU.
#
# ROB is free software; you can redistribute it and/or modify it under the
# terms of the MIT License; see LICENSE file for more details.

"""The reproducible benchmark engine uses a simple file store to maintain files
that are uploaded by users as part of the inputs to workflow runs.

Uploaded files are currently stored on disk. The file store maintains uploaded
files within sub-folders of a base folder. Each file has a unique identifier
that is generated by the system. The file handle is a wrapper around the local
file to provide access to the file.
"""

import os
from datetime import datetime

from benchtmpl.util.core import get_unique_identifier


class FileHandle(object):
    """Handle for files that are managed by the file store. Each file has a
    unique identifier and a file name. Files are maintaind in folders on the
    file system.
    """
    def __init__(self, filepath, identifier=None, file_name=None):
        """Initialize the file identifier, the (full) file path, and the file
        name. The file path is mandatory.

        Parameters
        ----------
        filepath: string
            Absolute path to file on disk
        identifier: string, optional
            Unique file identifier
        file_name: string, optional
            Base name of the file
        """
        self.filepath = os.path.abspath(filepath)
        self.identifier = identifier if not identifier is None else get_unique_identifier()
        self.file_name = file_name if not file_name is None else os.path.basename(self.filepath)

    @property
    def created_at(self):
        """Date and time when file was created (e.g., uploaded). The timestamp
        is given in the UTC time zone.

        Returns
        -------
        datetime.date
        """
        return datetime.utcfromtimestamp(os.path.getctime(self.filepath))

    @property
    def name(self):
        """Method for accessing the file name.

        Returns
        -------
        string
        """
        return self.file_name

    @property
    def size(self):
        """Get size of file in bytes.

        Returns
        -------
        int
        """
        return os.stat(self.filepath).st_size



# This file is part of the Reproducible Open Benchmarks for Data Analysis
# Platform (ROB).
#
# Copyright (C) 2019 NYU.
#
# ROB is free software; you can redistribute it and/or modify it under the
# terms of the MIT License; see LICENSE file for more details.

"""The template repository maintains workflow templates together with any static
files that are required to run a parameterized workflow.

The default repository implementation stores all template information in files
that are organized under a given base directory.
"""

import git
import os
import shutil

from benchtmpl.workflow.template.base import TemplateHandle
from benchtmpl.workflow.template.loader import DefaultTemplateLoader

import benchtmpl.error as err
import benchtmpl.util.core as util


"""Names for files and folders that are used to maintain template information."""
STATIC_FILES_DIR = 'static'
TEMPLATE_FILE = 'template.json'


""" "Default value for max. attempts parameters."""
DEFAULT_MAX_ATTEMPTS = 100


class TemplateRepository(object):
    """The template repository maintains a set of workflow templates. Each
    template is stored in a folder on the file system. The folder contains all
    static files that are provided when adding a new workflow template to the
    repository (i.e., files that are required to run the workflow).

    The repository can be used to store template handles of different types. It
    is the responsibility of the provided template loader to read and write
    template specification files to/from disk.
    """
    def __init__(
        self, base_dir,  loader=None, filenames=None, suffixes=None,
        id_func=None, max_attempts=DEFAULT_MAX_ATTEMPTS
    ):
        """Initialize the base directory where templates are maintained. The
        optional identifier function is used to generate unique template
        identifier. By default, short identifier are used.

        Uses the cross product of file names and suffixes to look for a
        template specification file when adding a new template.

        Parameters
        ----------
        base_dir: string
            Base directory for templates
        loader: benchtmpl.workflow.template.loader.TemplateLoader, optional
            Loader for reading and writing template specification files
        filenames: list(string)
            List of file names for templates files
        suffixes: list(string)
            List of recognized file suffies for template files
        id_func: func, optional
            Function to generate template folder identifier
        max_attempts: int, optional
            Maximum number of attempts to create a unique folder for a new
            workflow template
        """
        self.base_dir = os.path.abspath(base_dir)
        self.loader = loader if not loader is None else DefaultTemplateLoader()
        self.filenames = filenames if not filenames is None else ['template', 'workflow']
        self.suffixes = suffixes if not suffixes is None else ['.yml', '.yaml', '.json']
        self.id_func = id_func if not id_func is None else util.get_short_identifier
        self.max_attempts = max_attempts
        # Create the base directory if it does not exist
        util.create_dir(self.base_dir)

    def add_template(self, src_dir=None, src_repo_url=None, template_spec_file=None):
        """Create file and folder structure for a new workflow template. Assumes
        that either a workflow source directory or the Url of a remote Git
        repository is given.

        Creates a new folder with unique name in the base directory of the
        template store. The created folder will contain a copy of the source
        folder or the git repository.

        The source folder is expected to contain the template specification
        file. If the template_spec file is not given the method will look for a
        file using the entries in the file name and file suffix lists.

        If no template file is found in the source folder a ValueError is
        raised. The contents of the source directory will be copied to the
        new template directory (as subfolder named 'workflow'). The template
        directory may also contain other subfolders, e.g., subfolders for
        individual runs containing downloaded result files.

        Parameters
        ----------
        src_dir: string, optional
            Directory containing the workflow components, i.e., the fixed
            files and the template specification (optional).
        src_repo_url: string, optional
            Git repository that contains the the workflow components
        template_spec_file: string, optional
            Path to the workflow template specification file (absolute or
            relative to the workflow directory)

        Returns
        -------
        benchtmpl.workflow.template.TemplateHandle

        Raises
        ------
        benchtmpl.error.InvalidParameterError
        benchtmpl.error.InvalidTemplateError
        ValueError
        """
        # Exactly one of src_dir and src_repo_url has to be not None. If both
        # are None (or not None) a ValueError is raised.
        if src_dir is None and src_repo_url is None:
            raise ValueError('both \'src_dir\' and \'src_repo_url\' are missing')
        elif not src_dir is None and not src_repo_url is None:
            raise ValueError('cannot have both \'src_dir\' and \'src_repo_url\'')
        # Create a new unique folder for the template resources
        identifier = None
        template_dir = None
        attempt = 0
        while identifier is None or template_dir is None:
            identifier = self.id_func()
            template_dir = os.path.join(self.base_dir, identifier)
            if os.path.isdir(template_dir):
                identifier = None
                template_dir = None
                attempt += 1
                if attempt > self.max_attempts:
                    raise RuntimeError('could not create unique directory')
        # Create the new folder (this should be unique now)
        os.makedirs(template_dir)
        try:
            # Copy either the given workflow directory into the created template
            # folder or clone the Git repository.
            static_dir = os.path.join(template_dir, STATIC_FILES_DIR)
            if not src_dir is None:
                shutil.copytree(src=src_dir, dst=static_dir)
            else:
                git.Repo.clone_from(src_repo_url, static_dir)
            # Find template specification file in the template workflow folder.
            # If the file is not found the template directory is removed and a
            # ValueError is raised.
            candidates = list()
            for name in self.filenames:
                for suffix in self.suffixes:
                    candidates.append(os.path.join(static_dir, name + suffix))
            if not template_spec_file is None:
                candidates = [template_spec_file] + candidates
            for filename in candidates:
                if os.path.isfile(filename):
                    # Read template from file. If no error occurs the folder
                    # contains a valid template.
                    template = self.loader.load(
                        filename=filename,
                        identifier=identifier,
                        base_dir=static_dir,
                        validate=True
                    )
                    # Store serialized template handle on disk
                    self.loader.write(
                        template=template,
                        filename=os.path.join(template_dir, TEMPLATE_FILE)
                    )
                    return template
        except (IOError, OSError, ValueError, err.TemplateError) as ex:
            # Make sure to cleanup by removing the created template folder
            shutil.rmtree(template_dir)
            raise err.InvalidTemplateError(str(ex))
        # No template file found. Cleanup and raise error
        shutil.rmtree(template_dir)
        raise err.InvalidTemplateError('no template file found')

    def delete_template(self, identifier):
        """Delete all resources that are associated with the given template.
        This will delete the directory on disk that contains the template
        resources.

        The result is True if the template existed and False otherwise.

        Parameters
        ----------
        identifier: string
            Unique template identifier

        Returns
        -------
        bool

        Raises
        ------
        IOError
        OSError
        """
        # Drop the template directory if it exists
        template_dir = os.path.join(self.base_dir, identifier)
        if os.path.isdir(template_dir):
            shutil.rmtree(template_dir)
            return True
        return False

    def get_template(self, identifier):
        """Get handle for the template with the given identifier.

        Parameters
        ----------
        identifier: string
            Unique template identifier

        Returns
        -------
        benchtmpl.workflow.template.base.TemplateHandle

        Raises
        ------
        benchtmpl.error.InvalidTemplateError
        benchtmpl.error.UnknownTemplateError
        """
        # Raise exception if the template directory does not exist
        template_dir = os.path.join(self.base_dir, identifier)
        if not os.path.isdir(template_dir):
            raise err.UnknownTemplateError(identifier)
        return self.loader.load(
            filename=os.path.join(template_dir, TEMPLATE_FILE),
            base_dir=os.path.join(template_dir, STATIC_FILES_DIR)
        )



# This file is part of the Reproducible Open Benchmarks for Data Analysis
# Platform (ROB).
#
# Copyright (C) 2019 NYU.
#
# ROB is free software; you can redistribute it and/or modify it under the
# terms of the MIT License; see LICENSE file for more details.

"""Base classes and methods that are used for unit tests."""


class FakeStream(object):
    """Fake stream object to test upload from stream. Needs to implement the
    save(filename) method.
    """
    def save(self, filename):
        """Write simple text to given file."""
        with open(filename, 'w') as f:
            f.write('This is a fake\n')



"""Test various methods that copy input and output files for workflow runs."""

import os
import shutil

from unittest import TestCase

from benchtmpl.backend.util import FileCopy
from benchtmpl.io.files.base import FileHandle
from benchtmpl.workflow.parameter.value import TemplateArgument
from benchtmpl.workflow.template.base import TemplateHandle
from benchtmpl.workflow.template.repo import TemplateRepository

import benchtmpl.error as err
import benchtmpl.backend.util as backend


DATA_FILE = './tests/files/workflows/helloworld/data/names.txt'
TMP_DIR = './tests/files/.tmp'
INPUT_DIR = './tests/files/workflows/helloworld'
INPUT_FILE = './tests/files/schema.json'
SPEC_FILE = 'alt-template.yaml'
SPEC_FILE_ERR = 'alt-error.yaml'
WORKFLOW_DIR = './tests/files/template'


class TestFileCopy(TestCase):
    """Test copying files on local disk for workflow run preparation."""
    def setUp(self):
        """Create an empty target directory for each test and intitialize the
        file loader function.
        """
        self.tearDown()
        os.makedirs(TMP_DIR)
        self.loader = FileCopy(TMP_DIR)

    def tearDown(self):
        """Remove the temporary target directory."""
        if os.path.isdir(TMP_DIR):
            shutil.rmtree(TMP_DIR)

    def test_input_dir_copy(self):
        """Test copying local directories into a workflow run directory."""
        # Copy file to target directory
        self.loader(source=INPUT_DIR, target='workflow')
        dirname = os.path.join(TMP_DIR, 'workflow')
        self.assertTrue(os.path.isdir(dirname))
        self.assertTrue(os.path.isdir(os.path.join(dirname, 'code')))
        datadir = os.path.join(dirname, 'data')
        self.assertTrue(os.path.isdir(datadir))
        self.assertTrue(os.path.isfile(os.path.join(datadir, 'names.txt')))
        # Copy to target directory under parent that does not exist
        dst = os.path.join('run', 'files', 'wf')
        self.loader(source=INPUT_DIR, target=dst)
        dirname = os.path.join(TMP_DIR, dst)
        self.assertTrue(os.path.isdir(dirname))
        self.assertTrue(os.path.isdir(os.path.join(dirname, 'code')))
        datadir = os.path.join(dirname, 'data')
        self.assertTrue(os.path.isdir(datadir))
        self.assertTrue(os.path.isfile(os.path.join(datadir, 'names.txt')))

    def test_input_file_copy(self):
        """Test copying local input files into a workflow run directory."""
        # Copy file to target directory
        self.loader(source=INPUT_FILE, target='input.data')
        self.assertTrue(os.path.isfile(os.path.join(TMP_DIR, 'input.data')))
        # Copy file to non-existing target directory
        target = os.path.join('data', 'input.data')
        self.loader(source=INPUT_FILE, target=target)
        self.assertTrue(os.path.isfile(os.path.join(TMP_DIR, target)))

    def test_prepare_inputs_for_local_run(self):
        """Test copying input files for a local workflow run."""
        # Load template
        store = TemplateRepository(base_dir=TMP_DIR)
        template = store.add_template(
            src_dir=WORKFLOW_DIR,
            template_spec_file=os.path.join(WORKFLOW_DIR, SPEC_FILE)
        )
        # Create run directory
        run_dir = os.path.join(TMP_DIR, 'run')
        os.makedirs(run_dir)
        # Copy input files to run directory
        backend.upload_files(
            template=template,
            files=template.workflow_spec.get('inputs', {}).get('files', []),
            arguments={
                'names': TemplateArgument(
                    template.get_parameter('names'),
                    value=FileHandle(filepath=DATA_FILE)
                )
            },
            loader=FileCopy(run_dir)
        )
        # We should have the following files in the run directory:
        # code/helloworld.py
        # data/persons.txt
        # data/friends.txt
        self.assertTrue(os.path.isfile(os.path.join(run_dir, 'code', 'helloworld.py')))
        self.assertTrue(os.path.isfile(os.path.join(run_dir, 'data', 'persons.txt')))
        self.assertTrue(os.path.isfile(os.path.join(run_dir, 'data', 'friends.txt')))
        # data/persons.txt should contain Alice and Bob
        names = set()
        with open(os.path.join(run_dir, 'data', 'persons.txt'), 'r') as f:
            for line in f:
                names.add(line.strip())
        self.assertEqual(len(names), 2)
        self.assertTrue('Alice' in names)
        self.assertTrue('Bob' in names)
        # data/friends contains Jane Doe and Joe Bloggs
        friends = set()
        with open(os.path.join(run_dir, 'data', 'friends.txt'), 'r') as f:
            for line in f:
                friends.add(line.strip())
        self.assertEqual(len(friends), 2)
        self.assertTrue('Jane Doe' in friends)
        self.assertTrue('Joe Bloggs' in friends)
        # Error cases
        with self.assertRaises(err.MissingArgumentError):
            backend.upload_files(
                template=template,
                files=template.workflow_spec.get('inputs', {}).get('files', []),
                arguments={},
                loader=FileCopy(run_dir)
            )
        # Error when copying non-existing file
        template = store.add_template(
            src_dir=WORKFLOW_DIR,
            template_spec_file=os.path.join(WORKFLOW_DIR, SPEC_FILE)
        )
        shutil.rmtree(run_dir)
        os.makedirs(run_dir)
        with self.assertRaises(IOError):
            backend.upload_files(
                template=template,
                files=template.workflow_spec.get('inputs', {}).get('files', []),
                arguments={
                    'names': TemplateArgument(
                        template.get_parameter('names'),
                        value=FileHandle(filepath=os.path.join(TMP_DIR, 'no.file'))
                    )
                },
                loader=FileCopy(run_dir)
            )
        self.assertFalse(os.path.isdir(os.path.join(run_dir, 'data')))
        # If the constant value for the names parameter is removed the names
        # file is copied to the run directory and not to the data folder
        para = template.get_parameter('names')
        para.as_constant = None
        shutil.rmtree(run_dir)
        os.makedirs(run_dir)
        backend.upload_files(
            template=template,
            files=template.workflow_spec.get('inputs', {}).get('files', []),
            arguments={
                'names': TemplateArgument(
                    parameter=para,
                    value=FileHandle(filepath=DATA_FILE)
                )
            },
            loader=FileCopy(run_dir)
        )
        # We should have the following files in the run directory:
        # code/helloworld.py
        # names.txt
        # data/friends.txt
        self.assertTrue(os.path.isfile(os.path.join(run_dir, 'code', 'helloworld.py')))
        self.assertTrue(os.path.isfile(os.path.join(run_dir, 'names.txt')))
        self.assertFalse(os.path.isfile(os.path.join(run_dir, 'data', 'persons.txt')))
        self.assertTrue(os.path.isfile(os.path.join(run_dir, 'data', 'friends.txt')))
        # Template with input file parameter that is not of type file
        template = store.add_template(
            src_dir=WORKFLOW_DIR,
            template_spec_file=os.path.join(WORKFLOW_DIR, SPEC_FILE_ERR)
        )
        shutil.rmtree(run_dir)
        os.makedirs(run_dir)
        # Copy input files to run directory
        with self.assertRaises(err.InvalidTemplateError):
            backend.upload_files(
                template=template,
                files=template.workflow_spec.get('inputs', {}).get('files', []),
                arguments={
                    'sleeptime': TemplateArgument(
                        template.get_parameter('names'),
                        value=FileHandle(filepath=DATA_FILE)
                    )
                },
                loader=FileCopy(run_dir)
            )


if __name__ == '__main__':
    import unittest
    unittest.main()



# This file is part of the Reproducible Open Benchmarks for Data Analysis
# Platform (ROB).
#
# Copyright (C) 2019 NYU.
#
# ROB is free software; you can redistribute it and/or modify it under the
# terms of the MIT License; see LICENSE file for more details.

"""Abstract interface for the workflow execution engine. The reproducible
benchmark engine is flexible with respect to the workflow engine that is being
used.

Each engine has to implement the abstract workflow engine class that is defined
in this module. The engine is responsible for interpreting a given workflow
template and a set of template parameter arguments. The engine is also
responsible for retrieving output files and for providing access to these files.
"""

from abc import abstractmethod


class WorkflowEngine(object):
    """The workflow engine is used to execute workflow templates for a given
    set of arguments for template parameters as well as to check the state of
    the workflow execution.

    Workflow executions, referred to as runs, are identified by unique run ids
    that are assigned by the engine when the execution starts.
    """
    @abstractmethod
    def cancel_run(self, run_id):
        """Request to cancel execution of the given run.

        Parameters
        ----------
        run_id: string
            Unique run identifier
        """
        raise NotImplementedError()

    @abstractmethod
    def exec(self, template, arguments):
        """Execute a given workflow template for a set of argument values.
        Returns an unique identifier for the started workflow run.

        Parameters
        ----------
        template: benchtmpl.workflow.template.base.TemplateHandle
            Workflow template containing the parameterized specification and the
            parameter declarations
        arguments: dict(benchtmpl.workflow.parameter.value.TemplateArgument)
            Dictionary of argument values for parameters in the template

        Returns
        -------
        string
        """
        raise NotImplementedError()

    @abstractmethod
    def get_state(self, run_id):
        """Get the status of the workflow with the given identifier.

        Parameters
        ----------
        run_id: string
            Unique run identifier

        Returns
        -------
        benchtmpl.workflow.state.WorkflowState

        Raises
        ------
        benchtmpl.error.UnknownRunError
        """
        raise NotImplementedError()



# This file is part of the Reproducible Open Benchmarks for Data Analysis
# Platform (ROB).
#
# Copyright (C) 2019 NYU.
#
# ROB is free software; you can redistribute it and/or modify it under the
# terms of the MIT License; see LICENSE file for more details.

"""Extended workflow template for data analytics benchmarks. The benchmark
template extends the base template with one element that contains the
specification for the benchmark result.

The assumption is that the workflow for the benchmark contains one step that
generates a result file. This file is expected to contain a JSON or YAML object
with the summarized results for the benchmark workflow run. These results are
used to generate the overall benchmark leaderboard.

The structure of the benchmark result element is defined in the schema module.
"""

from benchtmpl.workflow.template.base import TemplateHandle


class BenchmarkTemplate(TemplateHandle):
    """Extended workflow template for data analytics benchmarks that add the
    result schema object to the base template.
    """
    def __init__(self, workflow_spec, schema, identifier=None, base_dir=None, parameters=None):
        """Initialize the components of the benchmark template. A super class
        raises an error if the identifier of template parameters are not unique.

        Parameters
        ----------
        workflow_spec: dict
            Workflow specification object
        schema: benchtmpl.workflow.benachmark.schema.BenchmarkResultSchema
            Schema of the result file that is generated for each workflow run
        identifier: string, optional
            Unique template identifier. If no value is given a UUID will be
            assigned.
        base_dir: string, optional
            Optional path to directory on disk that contains static files that
            are required to run the represented workflow
        parameters: list(benchtmpl.workflow.parameter.base.TemplateParameter), optional
            List of workflow template parameter declarations

        Raises
        ------
        benchtmpl.error.InvalidTemplateError
        """
        super(BenchmarkTemplate, self).__init__(
            workflow_spec=workflow_spec,
            identifier=identifier,
            base_dir=base_dir,
            parameters=parameters
        )
        self.schema = schema



from setuptools import setup

readme = open('README.rst').read()


install_requires=[
    'future',
    'gitpython',
    'jsonschema',
    'pyyaml>=5.1'
]


tests_require = [
    'coverage>=4.0',
    'coveralls',
    'nose'
]


extras_require = {
    'docs': [
        'Sphinx',
        'sphinx-rtd-theme'
    ],
    'tests': tests_require,
}


setup(
    name='benchmark-templates',
    version='0.1.1',
    description='Workflow Templates for Reproducible Data Analysis Benchmarks',
    long_description=readme,
    long_description_content_type='text/x-rst',
    keywords='reproducibility benchmarks data analysis',
    url='https://github.com/scailfin/benchmark-templates',
    author="Heiko Mueller",
    author_email="heiko.muller@gmail.com",
    license='MIT',
    packages=['benchtmpl'],
    include_package_data=True,
    test_suite='nose.collector',
    extras_require=extras_require,
    tests_require=tests_require,
    install_requires=install_requires
)



"""Test simple file store functionality."""

import datetime
import os
import shutil

from unittest import TestCase

from benchtmpl.io.files.store import Filestore
from benchtmpl.util.tests import FakeStream


LOCAL_FILE = 'tests/files/schema.json'
TMP_DIR = './tests/files/.tmp'


class TestFilestore(TestCase):
    def setUp(self):
        """Remove temp directory if it exists.."""
        if os.path.isdir(TMP_DIR):
            shutil.rmtree(TMP_DIR)

    def tearDown(self):
        """Remove temporary file store directory."""
        if os.path.isdir(TMP_DIR):
            shutil.rmtree(TMP_DIR)

    def test_create_and_delete_file(self):
        """Test creating a new run resource for a template handle."""
        fs = Filestore(directory=TMP_DIR)
        # Create new entry from local file
        fh = fs.upload_file(LOCAL_FILE)
        self.assertTrue(os.path.isdir(os.path.join(TMP_DIR, fh.identifier)))
        self.assertTrue(os.path.isfile(fh.filepath))
        self.assertEqual(fh.name, 'schema.json')
        self.assertEqual(fh.size, 1765)
        self.assertTrue(isinstance(fh.created_at, datetime.date))
        # Get the file handle
        fh1 = fs.get_file(fh.identifier)
        self.assertEqual(fh.identifier, fh1.identifier)
        self.assertEqual(fh.name, fh1.name)
        # Delete the uploaded file
        result = fs.delete_file(fh.identifier)
        self.assertTrue(result)
        self.assertFalse(os.path.isdir(os.path.join(TMP_DIR, fh.identifier)))
        self.assertFalse(os.path.isfile(fh.filepath))
        self.assertIsNone(fs.get_file(fh.identifier))
        # Deleting a non existing file returns False
        result = fs.delete_file(fh.identifier)
        self.assertFalse(result)

    def test_upload_get_and_list_files(self):
        """Test uploading and accessing files for a run."""
        fs = Filestore(directory=TMP_DIR)
        fh1 = fs.upload_file(LOCAL_FILE)
        fh2 = fs.upload_file(LOCAL_FILE)
        # Both files have the same name but different identifier (and point to
        # different paths)
        self.assertNotEqual(fh1.identifier, fh2.identifier)
        self.assertNotEqual(fh1.filepath, fh2.filepath)
        self.assertEqual(fh1.name, fh2.name)
        # File listing contains two elements
        files = fs.list_files()
        self.assertEqual(len(files), 2)
        self.assertEqual(files[0].name, files[1].name)
        self.assertNotEqual(files[0].identifier, files[1].identifier)
        self.assertNotEqual(files[0].filepath, files[1].filepath)
        # Recreate the file store and check that the two files still exist
        fs = Filestore(directory=TMP_DIR)
        files = fs.list_files()
        self.assertEqual(len(files), 2)
        self.assertEqual(files[0].name, files[1].name)
        self.assertNotEqual(files[0].identifier, files[1].identifier)
        self.assertNotEqual(files[0].filepath, files[1].filepath)
        # Unpload unknown file will raise value error
        with self.assertRaises(ValueError):
            fs.upload_file('tests/files/thisisnotafile.noway')
        # Upload third file from stream
        fh = fs.upload_stream(FakeStream(), 'file.txt')
        self.assertEqual(len(fs.list_files()), 3)
        self.assertEqual(fh.name, 'file.txt')
        # Delete file
        fs.delete_file(fh.identifier)
        files = fs.list_files()
        self.assertEqual(len(files), 2)
        self.assertFalse(fh.identifier in [f.identifier for f in files])
        # Recreate the file store and check that only the two files still exist
        fs = Filestore(directory=TMP_DIR)
        files = fs.list_files()
        self.assertEqual(len(files), 2)
        self.assertFalse(fh.identifier in [f.identifier for f in files])


if __name__ == '__main__':
    import unittest
    unittest.main()



# This file is part of the Reproducible Open Benchmarks for Data Analysis
# Platform (ROB).
#
# Copyright (C) 2019 NYU.
#
# ROB is free software; you can redistribute it and/or modify it under the
# terms of the MIT License; see LICENSE file for more details.
